#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 22:20:28 2018

@author: omartin
"""
#%% MANAGE PYTHON LIBRAIRIES
import numpy as np
import time
import sys
import numpy.fft as fft

from p3.aoSystem.aoSystem import aoSystem
from p3.aoSystem.fourierModel import fourierModel
import p3.aoSystem.FourierUtils as FourierUtils
import p3.psfr.psfrUtils as psfrUtils
from p3.aoSystem.frequencyDomain import frequencyDomain as frequencyDomain

#%%
rad2mas = 3600 * 180 * 1000 / np.pi
rad2arc = rad2mas / 1000

class psfR:
    """
    """
    # INIT
    def __init__(self,trs,path_root='',nLayer=None,theta_ext=0):
        """
        """
        # READ PARFILE
        tstart = time.time()

        # PARSING INPUTS
        if hasattr(trs,'path_ini') == False:
            print('ERROR : no .ini file attached with the telemetry object')
            return
        self.path_ini  = trs.path_ini
        self.trs       = trs
        self.theta_ext = theta_ext
        self.ao        = aoSystem(self.path_ini,path_root=path_root)
        self.tag       = 'PSF-R'

        # CHECK THE PUPIL
        if self.ao.tel.path_pupil == '' and np.any(trs.tel.pupil):
            self.ao.tel.pupil = self.trs.tel.pupil

        if self.ao.tel.path_static_on and hasattr(trs.tel,'map_ncpa')and np.any(trs.tel.map_ncpa):
            self.ao.tel.opdMap_on = trs.tel.map_ncpa

        if self.ao.error == False:

            # DEFINING THE FREQUENCY DOMAIN
            self.freq = frequencyDomain(self.ao)

            # DEFINING BOUNDS
            self.bounds = self.defineBounds()

            # INSTANTIATING THE FOURIER MODEL
            self.fao = fourierModel(self.path_ini,calcPSF=False,display=False)

            # INSTANTIATING THE FITTING PHASE STRUCTURE FUNCTION FOR r0=1m
            self.dphi_fit = self.fittingPhaseStructureFunction(1)

            # INSTANTIATING THE ALIASING PHASE STRUCTURE FUNCTION FOR r0=1m
            self.dphi_alias = self.aliasingPhaseStructureFunction(1)

            # INSTANTIATING THE AO RESIDUAL PHASE STRUCTURE FUNCTION
            self.dphi_ao = self.aoResidualStructureFunction()

            # INSTANTIATING THE TT RESIDUAL PHASE STRUCTURE FUNCTION IN LGS MODE
            # IN NGS MODE, THE TIP-TILT CONTRIBUTION SHOULD BE ADDED TO THE RESIDUAL WAVEFRONT
            # TO ACCOUNT FOR THE CROSS-TERMS
            self.dphi_tt = self.tipTiltPhaseStructureFunction()

            # INSTANTIATING THE ANISOPLANATISM PHASE STRUCTURE FUNCTION IF ANY
            if (self.ao.lgs==None) or (self.ao.lgs.height == 0):
                #self.dphi_ani = anisoplanatismStructureFunction(\
                #self.ao.tel,self.ao.atm,self.ao.src,self.ao.ngs,self.ao.ngs,\
                #self.freq.nOtf,self.freq.sampRef,self.ao.dms.nActu1D,Hfilter=1)#self.trs.mat.Hdm)
                if self.freq.isAniso:
                    self.dphi_ani = self.freq.dani_ang
            else:
                #self.dani_focang, self.dani_ang, self.dani_tt = anisoplanatism_structure_function(\
                #self.ao.tel,self.ao.atm,self.ao.src,self.ao.lgs,self.ao.ngs,\
                #self.freq.nOtf,self.freq.sampRef,self.ao.dms.nActu1D,Hfilter=1)#self.trs.mat.Hdm)

                self.dani_focang = self.freq.dani_focang
                self.dani_ang = self.freq.dani_ang
                self.dani_tt  = self.freq.dani_tt
                self.dphi_ani = self.dani_focang + self.dani_tt

            # COMPUTING THE DETECTOR PIXEL TRANSFER FUNCTION
            if self.trs.tel.name != 'KASP':
                self.otfPixel = self.pixelOpticalTransferFunction()
            else:
                self.otfPixel = 1.0

            # COMPUTING THE ERROR BREAKDOWN:
            self.get_error_breakdown()

        self.t_init = 1000*(time.time()  - tstart)

    def _repr__(self):
        return 'PSF-Reconstruction model'

    def defineBounds(self):
          #r0, gao, gtt, F , dx , dy , bg , stat
          _EPSILON = np.sqrt(sys.float_info.epsilon)

          # Bounds on r0
          bounds_down = list(np.ones(self.ao.atm.nL)*_EPSILON)
          bounds_up   = list(np.inf * np.ones(self.ao.atm.nL))
          # optical gains
          bounds_down += [0,0]
          bounds_up   += [np.inf,np.inf]
          # Photometry
          bounds_down += list(np.zeros(self.ao.src.nSrc))
          bounds_up   += list(np.inf*np.ones(self.ao.src.nSrc))
          # Astrometry
          bounds_down += list(-self.freq.nPix//2 * np.ones(2*self.ao.src.nSrc))
          bounds_up   += list( self.freq.nPix//2 * np.ones(2*self.ao.src.nSrc))
          # Background
          bounds_down += [-np.inf]
          bounds_up   += [np.inf]
          # Static aberrations
          bounds_down += list(-self.freq.wvlRef/2*1e9 * np.ones(self.ao.tel.nModes))
          bounds_up   += list(self.freq.wvlRef/2 *1e9 * np.ones(self.ao.tel.nModes))
          return (bounds_down,bounds_up)

    def fittingPhaseStructureFunction(self,r0):
        return r0**(-5/3) * np.real(fft.fftshift(FourierUtils.cov2sf(FourierUtils.psd2cov(self.freq.psdKolmo_,2*self.freq.kc_/self.freq.resAO))))

    def aliasingPhaseStructureFunction(self,r0):
        # computing the aliasing PSD over the AO-corrected area
        self.psdAlias_ = self.fao.aliasingPSD()/self.fao.ao.atm.r0**(-5/3)

        # zero-padding the PSD
        self.psdAlias_ = FourierUtils.enlargeSupport(self.psdAlias_,self.freq.nOtf/self.freq.resAO)

        # computing the aliasing phase structure function
        dphi_alias = r0**(-5/3) * np.real(fft.fftshift(FourierUtils.cov2sf(FourierUtils.psd2cov(self.psdAlias_,2*self.freq.kc_/self.freq.resAO))))

        # interpolating the phase structure function if required
        if dphi_alias.shape[0] != self.freq.nOtf:
            dphi_alias = FourierUtils.interpolateSupport(dphi_alias,self.freq.nOtf,kind='spline')
        return dphi_alias

    def aoResidualStructureFunction(self,method='slopes-based',basis='Vii'):
        """
        """
        # computing the empirical covariance matrix of the AO-residual OPD in the DM actuators domain
        if method == 'slopes-based':
            du = self.trs.rec.res
        elif method == 'dm-based':
            du = np.diff(self.trs.dm.com,axis=0)/self.ao.rtc.holoop['gain']
        self.Cao  =  np.matmul(du.T,du)/du.shape[0]

        # Unbiasing noise and accounting for the wave number
        Cao = (2*np.pi/self.freq.wvlRef)**2 * (self.Cao - self.trs.wfs.Cn_ao)

        # Computing the phase structure function
        _,dphi_ao = psfrUtils.modes2Otf(Cao,self.trs.mat.dmIF,self.ao.tel.pupil,self.freq.nOtf,basis=basis,samp=self.freq.sampRef/2)

        return dphi_ao

    def tipTiltPhaseStructureFunction(self):
        """
        """
        # computing the empirical covariance matrix of the residual tip-tilt in meter
        self.Ctt = np.dot(self.trs.tipTilt.slopes.T,self.trs.tipTilt.slopes)/self.trs.tipTilt.nExp

        # computing the coefficients of the Gaussian Kernel in rad^2
        Guu = (2*np.pi/self.freq.wvlRef)**2 *(self.Ctt - self.trs.tipTilt.Cn_tt)

        # rotating the axes
        ang = self.trs.tel.pupilAngle * np.pi/180
        # freq.U_ ranges within [-1,1]; factor 2 to account for the TT normalization
        Ur  = 2*(self.freq.U_*np.cos(ang) + self.freq.V_*np.sin(ang))
        Vr  = 2*(-self.freq.U_*np.sin(ang) + self.freq.V_*np.cos(ang))

        # computing the Gaussian-Kernel
        dphi_tt = Guu[0,0]*Ur**2 + Guu[1,1]*Vr**2 + Guu[0,1]*Ur*Vr.T + Guu[1,0]*Vr*Ur.T

        return dphi_tt


    def pixelOpticalTransferFunction(self):
        """
        """
        #note : self.U_/V_ ranges ar -1 to 1
        otfPixel = np.sinc(self.freq.U_)* np.sinc(self.freq.V_)
        return otfPixel

    def TotalPhaseStructureFunction(self,r0,gho,gtt,Cn2=[]):

        # On-axis phase structure function
        SF   = gho*self.dphi_ao + gtt*self.dphi_tt + r0**(-5/3) * (self.dphi_fit + self.dphi_alias)

        # Anisoplanatism phase structure function
        if self.freq.isAniso and (len(Cn2) == self.freq.dani_ang.shape[1]):
            SF = SF[:,:,np.newaxis] + (self.dphi_ani.transpose(2,3,1,0) * Cn2[:,np.newaxis]).sum(axis=2)
        else:
            SF = np.repeat(SF[:,:,np.newaxis],self.ao.src.nSrc,axis=2)

        return SF/(2*np.pi*1e-9/self.freq.wvlRef)**2

    def __call__(self,x0,nPix=None):

        # ----------------- GETTING THE PARAMETERS
        # Cn2 profile
        nL   = self.ao.atm.nL
        if nL > 1: # fit the Cn2 profile
            Cn2  = np.asarray(x0[0:nL])
            r0   = np.sum(Cn2)**(-3/5)
        else: #fit the r0
            Cn2= []
            r0 = x0[0]

        # PSD
        gho = x0[nL]
        gtt = x0[nL+1]

        # Astrometry/Photometry/Background
        x0_stellar = np.array(x0[nL+2:nL+4+3*self.ao.src.nSrc])
        if len(x0_stellar):
            F  = x0_stellar[0:self.ao.src.nSrc][:,np.newaxis] * np.array(self.ao.cam.transmittance)[np.newaxis,:]
            dx = x0_stellar[self.ao.src.nSrc:2*self.ao.src.nSrc][:,np.newaxis] + np.array(self.ao.cam.dispersion[0])[np.newaxis,:]
            dy = x0_stellar[2*self.ao.src.nSrc:3*self.ao.src.nSrc][:,np.newaxis] + np.array(self.ao.cam.dispersion[1])[np.newaxis,:]
            bkg= x0_stellar[3*self.ao.src.nSrc]
        else:
            F  = np.repeat(np.array(self.ao.cam.transmittance)[np.newaxis,:]* np.ones(self.freq.nWvl),self.ao.src.nSrc,axis=0)
            dx = np.repeat(np.array(self.ao.cam.dispersion[0])[np.newaxis,:]* np.ones(self.freq.nWvl),self.ao.src.nSrc,axis=0)
            dy = np.repeat(np.array(self.ao.cam.dispersion[1])[np.newaxis,:]* np.ones(self.freq.nWvl),self.ao.src.nSrc,axis=0)
            bkg= 0.0

        # Static aberrations
        if len(x0) > nL + 2 + 3*self.ao.src.nSrc + 1:
            x0_stat = list(x0[nL+3+3*self.ao.src.nSrc:])
        else:
            x0_stat = []

        # ----------------- GETTING THE PHASE STRUCTURE FUNCTION
        self.SF = self.TotalPhaseStructureFunction(r0,gho,gtt,Cn2=Cn2)

        # ----------------- COMPUTING THE PSF
        PSF, self.SR = FourierUtils.SF2PSF(self.SF,self.freq,self.ao,\
                        F=F,dx=dx,dy=dy,bkg=bkg,nPix=nPix,xStat=x0_stat,otfPixel=self.otfPixel)
        return PSF

    def get_error_breakdown(self,r0=None,gho=1,gtt=1):
        '''
        Computing the AO error breakdown from the variance of each individual covariance terms.
        INPUTS:
            - an instance of the psfr object
        OUTPUTS
        '''
        sr2fwe = lambda x: np.sqrt(-np.log(x))* self.freq.wvlRef*1e9/2/np.pi
        otf_dl = self.freq.otfDL
        S      = otf_dl.sum()
        self.wfe = dict()

        #1. STATIC ABERRATIONS
        if np.any(self.ao.tel.opdMap_on):
            self.wfe['NCPA'] = np.std(self.ao.tel.opdMap_on[self.ao.tel.pupil.astype(bool)])
        else:
            self.wfe['NCPA'] = 0

        #2. DM FITTING ERROR
        if not r0:
            r0 = self.ao.atm.r0 * (self.freq.wvlRef/self.ao.atm.wvl)**1.2
        otf_fit = np.exp(-0.5 * r0**(-5/3) * self.dphi_fit)
        sr_fit  = np.sum(otf_dl * otf_fit)/S
        self.wfe['FITTING'] = sr2fwe(sr_fit)

        #3. WFS ALIASING ERROR
        otf_alias = np.exp(-0.5 * r0**(-5/3) * self.dphi_alias)
        sr_alias  = np.sum(otf_dl * otf_alias)/S
        self.wfe['ALIASING'] = sr2fwe(sr_alias)

        #4. WFS NOISE ERROR
        # noise on high-order modes
        msk = self.trs.dm.validActuators.reshape(-1)
        self.wfe['HO NOISE'] = 1e9 * np.sqrt(self.trs.holoop.tf.pn * np.mean(self.trs.wfs.Cn_ao[msk,msk]))
        # noise on tip-tilt modes
        self.wfe['TT NOISE'] = 1e9 * np.sqrt(self.trs.ttloop.tf.pn * np.diag(self.trs.tipTilt.Cn_tt).sum())

        #5. AO BANDWIDTH ERROR
        C = self.Cao - (1+self.trs.holoop.tf.pn)*self.trs.wfs.Cn_ao
        self.wfe['SERVO-LAG'] = 1e9*np.sqrt(np.mean(np.mean(C[msk,msk])))

        #6. RESIDUAL TIP-TILT
        C = np.diag(self.Ctt - (1+self.trs.ttloop.tf.pn)*self.trs.tipTilt.Cn_tt)
        self.wfe['TIP-TILT'] = np.sqrt(np.sum(C))*1e9
        #sr_tt                = np.sum( otf_dl * np.exp(-0.5*self.dphi_tt))/S
        #self.wfe['TIP-TILT'] = sr2fwe(sr_tt)

        #7. PIXEL TF
        sr_pixel = np.sum(otf_dl * self.otfPixel)/S
        self.wfe['PIXEL TF'] = np.sqrt(-np.log(sr_pixel))* self.freq.wvlRef*1e9/2/np.pi

        #8. ANISOPLANATISM
        if self.freq.isAniso:
            Cn2     = self.ao.atm.weights * self.ao.atm.r0**(-5/3) * (self.ao.atm.wvl/self.ao.src.wvl[0])**2
            dani    = (self.dphi_ani[0].transpose(1,2,0) * Cn2).sum(axis=2)
            otf_ani = np.exp(-0.5 * dani)
            sr_ani  = np.sum(otf_dl * otf_ani)/S
            self.wfe['TOTAL ANISOPLANATISM'] = sr2fwe(sr_ani)
        else:
            self.wfe['TOTAL ANISOPLANATISM'] = 0

        if self.trs.aoMode == 'LGS' and self.freq.isAniso:
            # Angular
            dani    = (self.dani_ang[0].transpose(1,2,0) * Cn2).sum(axis=2)
            otf_ani = np.exp(-0.5 * dani)
            sr_ani  = np.sum(otf_dl * otf_ani)/S
            self.wfe['ANGULAR ANISOPLANATISM'] = sr2fwe(sr_ani)
            #tiptilt
            dani    = (self.dani_tt[0].transpose(1,2,0) * Cn2).sum(axis=2)
            otf_ani = np.exp(-0.5 * dani)
            sr_ani  = np.sum(otf_dl * otf_ani)/S
            self.wfe['ANISOKINETISM'] = sr2fwe(sr_ani)
            # focal
            dani    = (self.dani_focang[0].transpose(1,2,0) * Cn2).sum(axis=2)
            otf_ani = np.exp(-0.5 * dani)
            sr_ani  = np.sum(otf_dl * otf_ani)/S
            self.wfe['FOCAL ANISOPLANATISM'] = np.sqrt(sr2fwe(sr_ani) **2 - self.wfe['ANGULAR ANISOPLANATISM']**2)
        else:
            self.wfe['ANGULAR ANISOPLANATISM'] = 0
            self.wfe['ANISOKINETISM']          = 0
            self.wfe['FOCAL ANISOPLANATISM']   = 0

        #9. TOTAL WFE
        self.wfe['TOTAL WFE'] =  np.sqrt(self.wfe['NCPA']**2 +  self.wfe['FITTING']**2
                + self.wfe['HO NOISE']**2 + self.wfe['TT NOISE']**2 + self.wfe['SERVO-LAG']**2
                + self.wfe['TIP-TILT']**2 + self.wfe['TOTAL ANISOPLANATISM']**2)

        self.wfe['TOTAL WFE WITH PIXEL'] = np.hypot(self.wfe['TOTAL WFE'],self.wfe['PIXEL TF'])

        #10. TOTAL STREHL-RATIO
        self.wfe['REF WAVELENGTH'] = self.freq.wvlRef
        self.wfe['MARECHAL SR'] = 1e2*np.exp(-(self.wfe['TOTAL WFE'] * 2*np.pi*1e-9/self.freq.wvlRef)**2 )
        self.wfe['MARECHAL SR WITH PIXEL'] = 1e2*np.exp(-(self.wfe['TOTAL WFE WITH PIXEL'] * 2*np.pi*1e-9/self.freq.wvlRef)**2 )

        if hasattr(self,'SR'):
            self.wfe['PSF SR'] = self.SR[0]
        self.wfe['IMAGE-BASED SR'] = 1e2*FourierUtils.getStrehl(self.trs.cam.image,self.ao.tel.pupil,self.freq.sampRef,method='max')


















#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr  5 14:42:49 2021

@author: omartin
"""

# IMPORTING PYTHON LIBRAIRIES
import os.path as ospath
import pathlib
from configparser import ConfigParser
import numpy as np

import yaml

# IMPORTING P3 MODULES
from p3.aoSystem.telescope import telescope
from p3.aoSystem.atmosphere import atmosphere
from p3.aoSystem.source import source
from p3.aoSystem.deformableMirror import deformableMirror
from p3.aoSystem.detector import detector
from p3.aoSystem.sensor import sensor
from p3.aoSystem.rtc import rtc
import p3.aoSystem.anisoplanatismModel as anisoplanatismModel

#%%
class aoSystem():

    def raiseMissingRequiredOpt(self,sec,opt):
        raise ValueError("'{}' is missing from section '{}'"
                         .format(opt,sec))
        
    def raiseMissingRequiredSec(self,sec):
        raise ValueError("The section '{}' is missing from the parameter file"
                         .format(sec))
        
    def raiseNotSameLength(self,sec,opt):
        raise ValueError("'{}' in section '{}' must have the same length"
                         .format(*opt,sec))

    def check_section_key(self, primary):        
        return primary in self.my_data_map.keys()
    
    def check_config_key(self, primary, secondary):
        if primary in self.my_data_map.keys():
            return secondary in self.my_data_map[primary].keys()
        else:
            return False

    def get_config_value(self, primary, secondary):
        return self.my_data_map[primary][secondary]
    
    def __init__(self,path_config,path_root='',nLayer=None,getPSDatNGSpositions=False,coo_stars=None):

        self.coo_stars = coo_stars
        self.error = False
        # verify if the file exists
        if ospath.isfile(path_config) == False:
            raise ValueError('The parameter file (.ini or .yml) could not be found.')
                
        if path_config[-4::]=='.ini':
            # open the .ini file
            config = ConfigParser()
            config.optionxform = str
            config.read(path_config)
            self.my_data_map = {} 
            for section in config.sections():
                self.my_data_map[section] = {}
                for name,value in config.items(section):
                    self.my_data_map[section].update({name:eval(value)})            
            
        elif path_config[-4::]=='.yml':
            with open(path_config) as f:
                my_yaml_dict = yaml.safe_load(f)        
            self.my_data_map = my_yaml_dict

        self.getPSDatNGSpositions = getPSDatNGSpositions
        
        #%% TELESCOPE
        #----- grabbing main parameters
        if not(self.check_section_key('telescope')):
            self.raiseMissingRequiredSec('telescope')
        
        if self.check_config_key('telescope','TelescopeDiameter'):
            self.D = self.get_config_value('telescope','TelescopeDiameter')
        else:
            self.raiseMissingRequiredOpt('telescope','TelescopeDiameter')
        
        if self.check_config_key('telescope','ZenithAngle'):
            zenithAngle = self.get_config_value('telescope','ZenithAngle')
            if zenithAngle >= 90:
                raise ValueError("'zenithAngle' [deg] cannot be greater than or equal to 90.")
        else:
            zenithAngle = 0.0
        
        airmass = 1/np.cos(zenithAngle*np.pi/180)
        
        if self.check_config_key('telescope','ObscurationRatio'):
            obsRatio = self.get_config_value('telescope','ObscurationRatio')
        else:
            obsRatio = 0.0
        
        if self.check_config_key('telescope','Resolution'):
            nPup = self.get_config_value('telescope','Resolution')
        else:
            self.raiseMissingRequiredOpt('telescope','Resolution')
            
        path_p3 = str(pathlib.Path(__file__).parent.parent.absolute())
            
        #----- PUPIL
        if self.check_config_key('telescope','PathPupil'):
            PathPupil = self.get_config_value('telescope','PathPupil')
            if path_root == '' and PathPupil[0:9]=='/aoSystem' :
                path_pupil = path_p3 + PathPupil
            elif path_root == '' and PathPupil[0:8]=='aoSystem' :
                path_pupil = path_p3 +'/'+ PathPupil
            else:
                path_pupil = path_root + PathPupil
        else:
            path_pupil = ''
                  
        if self.check_config_key('telescope','PupilAngle'):
            pupilAngle = self.get_config_value('telescope','PupilAngle')
        else:
            pupilAngle = 0.0
        
        if self.check_config_key('telescope','PathStaticOn'):
            PathStaticOn = self.get_config_value('telescope','PathStaticOn')
            if path_root == '' and PathStaticOn[0:9]=='/aoSystem' :
                path_static_on = path_p3 + PathStaticOn
            elif path_root == '' and PathStaticOn[0:8]=='aoSystem' :
                path_static_on = path_p3 +'/'+ PathStaticOn
            else:
                path_static_on = path_root + PathStaticOn
        else:
            path_static_on = None       
        
        if self.check_config_key('telescope','PathStaticOff'):
            PathStaticOff = self.get_config_value('telescope','PathStaticOff')
            if path_root == '' and PathStaticOff[0:9]=='/aoSystem' :
                path_static_off = path_p3 + PathStaticOff
            elif path_root == '' and PathStaticOff[0:8]=='aoSystem' :
                path_static_off = path_p3 +'/'+ PathStaticOff
            else:
                path_static_off = path_root + PathStaticOff
        else:
            path_static_off = None
        
        if self.check_config_key('telescope','PathStaticPos'):
            PathStaticPos = self.get_config_value('telescope','PathStaticPos')
            if path_root == '' and PathStaticPos[0:9]=='/aoSystem' :
                path_static_pos = path_p3 + PathStaticPos
            elif path_root == '' and PathStaticPos[0:8]=='aoSystem' :
                path_static_pos = path_p3 +'/'+ PathStaticPos
            else:
                path_static_pos = path_root + PathStaticPos
        else:
            path_static_pos = None
            
        #----- APODIZER
        if self.check_config_key('telescope','PathApodizer'):
            PathApodizer = self.get_config_value('telescope','PathApodizer')
            if path_root == '' and PathApodizer[0:9]=='/aoSystem' :
                path_apodizer = path_p3 + PathApodizer
            elif path_root == '' and PathApodizer[0:8]=='aoSystem' :
                path_apodizer = path_p3 +'/'+ PathApodizer
            else:
                path_apodizer = path_root + PathApodizer
        else:
            path_apodizer = ''
                
        #----- TELESCOPE ABERRATIONS
        if self.check_config_key('telescope', 'PathStatModes'):
            PathStatModes = self.get_config_value('telescope','PathStatModes')
            if path_root == '' and PathStatModes[0:9]=='/aoSystem' :
                path_statModes = path_p3 + PathStatModes
            elif path_root == '' and PathStatModes[0:8]=='aoSystem' :
                path_statModes = path_p3 +'/'+ PathStatModes
            else:
                path_statModes = path_root + PathStatModes
        else:
            path_statModes = ''
            
        #----- EXTRA ERROR
        if self.check_config_key('telescope', 'extraErrorNm'):       
            extraErrorNm = self.get_config_value('telescope','extraErrorNm')
        else:
            extraErrorNm = 0
            
        if self.check_config_key('telescope', 'extraErrorExp'):
            extraErrorExp = self.get_config_value('telescope','extraErrorExp')
        else:
            extraErrorExp = -2
            
        if self.check_config_key('telescope', 'extraErrorMin'):
            extraErrorMin = self.get_config_value('telescope','extraErrorMin')
        else:
            extraErrorMin = 0
            
        if self.check_config_key('telescope', 'extraErrorMax'):
            extraErrorMax = self.get_config_value('telescope','extraErrorMax')
        else:
            extraErrorMax = 0
            
        if self.check_config_key('telescope', 'extraErrorLoNm'):       
            extraErrorLoNm = self.get_config_value('telescope','extraErrorLoNm')
        else:
            extraErrorLoNm = extraErrorNm  
            
        if self.check_config_key('telescope', 'extraErrorLoExp'):
            extraErrorLoExp = self.get_config_value('telescope','extraErrorLoExp')
        else:
            extraErrorLoExp = extraErrorExp
            
        if self.check_config_key('telescope', 'extraErrorLoMin'):
            extraErrorLoMin = self.get_config_value('telescope','extraErrorLoMin')
        else:
            extraErrorLoMin = 0
            
        if self.check_config_key('telescope', 'extraErrorLoMax'):
            extraErrorLoMax = self.get_config_value('telescope','extraErrorLoMax')
        else:
            extraErrorLoMax = 0

        # ----- class definition     
        self.tel = telescope(self.D, nPup,
                             zenith_angle=zenithAngle,
                             obsRatio=obsRatio,
                             pupilAngle=pupilAngle,
                             path_pupil=path_pupil,
                             path_static_on=path_static_on,
                             path_static_off=path_static_off,
                             path_static_pos=path_static_pos,
                             path_apodizer=path_apodizer,
                             path_statModes=path_statModes,
                             extraErrorNm=extraErrorNm,
                             extraErrorExp=extraErrorExp,
                             extraErrorMin=extraErrorMin,
                             extraErrorMax=extraErrorMax,
                             extraErrorLoNm=extraErrorLoNm,
                             extraErrorLoExp=extraErrorLoExp,
                             extraErrorLoMin=extraErrorLoMin,
                             extraErrorLoMax=extraErrorLoMax)                     

        if self.check_config_key('telescope', 'TechnicalFoV'):
            self.TechnicalFoV = self.get_config_value('telescope','TechnicalFoV')
        else:
            self.TechnicalFoV = 0
            
        if self.check_config_key('telescope', 'windPsdFile'):
            self.windPsdFile = self.get_config_value('telescope','windPsdFile')
        else:
            self.windPsdFile = 0

        #%% ATMOSPHERE
        if not(self.check_section_key('atmosphere')):
            self.raiseMissingRequiredSec('atmosphere')
            
        if self.check_config_key('atmosphere','Wavelength'):
            wvlAtm = self.get_config_value('atmosphere','Wavelength') 
        else:
            wvlAtm = 500e-9
        
        if self.check_config_key('atmosphere','Seeing'):
            r0 = 0.976*wvlAtm/self.get_config_value('atmosphere','Seeing')*3600*180/np.pi 
        else:
            if self.check_config_key('atmosphere','r0_value'):
                r0 = self.get_config_value('atmosphere','r0_value') 
            else:
                self.raiseMissingRequiredOpt('atmosphere','Seeing')
        
        if self.check_config_key('atmosphere','L0'):
            L0 = self.get_config_value('atmosphere','L0') 
        else:
            L0 = 25
            
        if self.check_config_key('atmosphere','Cn2Weights'):
            weights = self.get_config_value('atmosphere','Cn2Weights') 
            if np.abs(np.sum(weights) - 1) > 1e-3:
                raise ValueError("Sum of 'Cn2Weights' in section 'atmosphere' is not equal to 1")
        else:
            weights = [1.0]
        
        if self.check_config_key('atmosphere','Cn2Heights'):
            heights = self.get_config_value('atmosphere','Cn2Heights') 
        else:
            heights = [0.0]
            
        if self.check_config_key('atmosphere','WindSpeed'):
            wSpeed = self.get_config_value('atmosphere','WindSpeed') 
        else:
            wSpeed = [10.0]
            
        if self.check_config_key('atmosphere','WindDirection'):
            wDir = self.get_config_value('atmosphere','WindDirection') 
        else:
            wDir = [0.0]

        #-----  verification
        if not (len(weights) == len(heights) == len(wSpeed) == len(wDir)):
            self.raiseNotSameLength('atmosphere',['Cn2Weights','Cn2Heights','WindSpeed','WindDirection'])

        #----- class definition
        self.atm = atmosphere(wvlAtm, r0*airmass**(-3.0/5.0),
                              weights,
                              np.array(heights)*airmass,
                              wSpeed,
                              wDir,
                              L0)            
        
        #%%  GUIDE STARS 
        if not(self.check_section_key('sources_HO')):
            self.raiseMissingRequiredSec('sources_HO')
            
        if self.check_config_key('sources_HO','Wavelength'):
            self.wvlGs     = np.unique(np.array(self.get_config_value('sources_HO','Wavelength')))
        else:
            self.raiseMissingRequiredOpt('sources_HO', 'Wavelength')
            return 0
        
        if self.check_config_key('sources_HO','Zenith'):
            self.zenithGs = self.get_config_value('sources_HO','Zenith') 
        else:
            self.zenithGs = [0.0]
            
        if self.check_config_key('sources_HO','Azimuth'):
            self.azimuthGs = self.get_config_value('sources_HO','Azimuth') 
        else:
            self.azimuthGs = [0.0]
            
        if self.check_config_key('sources_HO','Height'):
            heightGs  = self.get_config_value('sources_HO','Height') 
        else:
            heightGs = 0.0
                         
        # ----- verification
        if len(self.zenithGs) != len(self.azimuthGs):
            self.raiseNotSameLength('sources_HO', ['Azimuth','Zenith'])

        self.ngs = None
        self.lgs = None
        # ----- creating the source class
        if heightGs == 0 and not self.check_section_key('sources_LO'):
            self.ngs = source(self.wvlGs,
                              self.zenithGs,self.azimuthGs,
                              tag="NGS",verbose=True)
        else:
            self.lgs = source(self.wvlGs,
                              self.zenithGs,self.azimuthGs,
                              height=heightGs*airmass,
                              tag="LGS",verbose=True)

        self.configLO()
        self.configLO_SC()

    def configLO_SC(self):
        #%%  SCIENCE SOURCES
        if not(self.check_section_key('sources_science')):
            self.raiseMissingRequiredSec('sources_science') 
            
        if self.check_config_key('sources_science','Wavelength'):
            wvlSrc = np.array(self.get_config_value('sources_science','Wavelength'))
        else:
            self.raiseMissingRequiredOpt('sources_science', 'Wavelength')
        
        if self.check_config_key('sources_science','Zenith'):
            zenithSrc = self.get_config_value('sources_science','Zenith') 
        else:
            zenithSrc = [0.0]
            
        if self.check_config_key('sources_science','Azimuth'):
            azimuthSrc = self.get_config_value('sources_science','Azimuth') 
        else:
            azimuthSrc = [0.0]
        
        if np.any(self.coo_stars):
            zenithSrc = np.hypot(self.coo_stars[0],self.coo_stars[1])
            azimuthSrc = np.arctan2(self.coo_stars[0],self.coo_stars[1])
        
        #----- verification
        if len(zenithSrc) != len(azimuthSrc):
            self.raiseNotSameLength('sources_science', ['Zenith','Azimuth'])
            return

        if self.getPSDatNGSpositions and self.check_config_key('sources_LO','Wavelength'):
            zenithSrc = zenithSrc +  (self.get_config_value('sources_LO','Zenith'))
            azimuthSrc = azimuthSrc + (self.get_config_value('sources_LO','Azimuth'))
            
        #----- class definition
        self.src = source(wvlSrc,
                          zenithSrc, azimuthSrc,
                          tag="SCIENCE",verbose=True)   

        #%% HIGH-ORDER WAVEFRONT SENSOR
        if not(self.check_section_key('sensor_HO')):
            self.raiseMissingRequiredSec('sensor_HO')
            
        if self.check_config_key('sensor_HO','PixelScale'):
            psInMas = self.get_config_value('sensor_HO','PixelScale')
        else:
            self.raiseMissingRequiredOpt('sensor_HO', 'PixelScale')
        
        if self.check_config_key('sensor_HO','FieldOfView'):
            fov = self.get_config_value('sensor_HO','FieldOfView')
        else:
            self.raiseMissingRequiredOpt('sensor_HO', 'FieldOfView') 
        
        if self.check_config_key('sensor_HO','Binning'):
            Binning = self.get_config_value('sensor_HO','Binning')
        else:
            Binning = 1
            
        if self.check_config_key('sensor_HO','SpotFWHM'):
            spotFWHM = self.get_config_value('sensor_HO','SpotFWHM')
        else:
            spotFWHM = [[0.0, 0.0]]
            
        if self.check_config_key('sensor_HO','NumberPhotons'):
            nphHO = self.get_config_value('sensor_HO','NumberPhotons')
        else:
            nphHO = [np.inf]
        
        if self.check_config_key('sensor_HO','SigmaRON'):
            ron = self.get_config_value('sensor_HO','SigmaRON')
        else:
            ron = 0.0
        
        if self.check_config_key('sensor_HO','Gain'):
            self.detectorGainHO = self.get_config_value('sensor_HO','Gain')
        else:
            self.detectorGainHO = 1
            
        if self.check_config_key('sensor_HO','SkyBackground'):
            sky = self.get_config_value('sensor_HO','SkyBackground')
        else:
            sky = 0.0
        
        if self.check_config_key('sensor_HO','Dark'):
            dark = self.get_config_value('sensor_HO','Dark')
        else:
            dark = 0.0
            
        if self.check_config_key('sensor_HO','ExcessNoiseFactor'):
            excess = self.get_config_value('sensor_HO','ExcessNoiseFactor')
        else:
            excess = 1.0
         
        if self.check_config_key('sensor_HO','SpectralBandwidth'):
            bw = self.get_config_value('sensor_HO','SpectralBandwidth')
        else:
            bw = 0.0
        if self.check_config_key('sensor_HO','Transmittance'):
            tr = self.get_config_value('sensor_HO','Transmittance')
        else:
            tr = [1.0]
        if self.check_config_key('sensor_HO','Dispersion'):
            disp = self.get_config_value('sensor_HO','Dispersion')
        else:
            disp = [[0.0], [0.0]]
                
        if self.check_config_key('sensor_HO','WfsType'):
            wfstype = self.get_config_value('sensor_HO','WfsType')
        else:
            wfstype = 'Shack-Hartmann'
            
        if self.check_config_key('sensor_HO','NumberLenslets'):
            nL = self.get_config_value('sensor_HO','NumberLenslets')
        else:
            nL = [20]
            
        if self.check_config_key('sensor_HO','SizeLenslets'):
            dsub = self.get_config_value('sensor_HO','SizeLenslets')
        else:
            dsub = list(self.D/np.array(nL))
            
        if self.check_config_key('sensor_HO','Modulation'):
            modu = self.get_config_value('sensor_HO','Modulation')
        else:
            modu = None
            
        if self.check_config_key('sensor_HO','NoiseVariance'):
            NoiseVar = np.array(self.get_config_value('sensor_HO','NoiseVariance'))
        else:
            NoiseVar = [None]
            
        if self.check_config_key('sensor_HO','Algorithm'):
            algorithm = self.get_config_value('sensor_HO','Algorithm')
        else:
            algorithm = 'wcog'
            
        if self.check_config_key('sensor_HO','WindowRadiusWCoG'):
            wr = self.get_config_value('sensor_HO','WindowRadiusWCoG')
        else:
            wr = 5.0
            
        if self.check_config_key('sensor_HO','ThresholdWCoG = 0.0'):
            thr = self.get_config_value('sensor_HO','ThresholdWCoG = 0.0')
        else:
            thr = 0.0
            
        if self.check_config_key('sensor_HO','NewValueThrPix'):
            nv = self.get_config_value('sensor_HO','NewValueThrPix')
        else:
            nv = 0.0
         
        if self.check_config_key('sensor_HO','ExcessNoiseFactor'):
            excess = self.get_config_value('sensor_HO','ExcessNoiseFactor')
        else:
            excess = 1.0
            
        if self.check_config_key('sensor_HO','addMcaoWFsensConeError'):
            self.addMcaoWFsensConeError = self.get_config_value('sensor_HO','addMcaoWFsensConeError')
        else:
            self.addMcaoWFsensConeError = False

        self.wfs = sensor(psInMas, fov,
                          binning=Binning, spotFWHM=spotFWHM,
                          nph=nphHO, bandwidth=bw, transmittance=tr, dispersion=disp,
                          gain=self.detectorGainHO, ron=ron, sky=sky, dark=dark, excess=excess,
                          nL=nL, dsub=dsub, wfstype=wfstype, modulation=modu,
                          noiseVar=NoiseVar, algorithm=algorithm,
                          algo_param=[wr,thr,nv], tag="HO WFS")
#%% TIP-TILT SENSORS
        if self.check_section_key('sensor_LO'):
            self.configLOsensor()
        else:
            self.tts = None

 #%% REAL-TIME-COMPUTER
     
        if self.check_config_key('RTC','LoopGain_HO'):
            self.LoopGain_HO = self.get_config_value('RTC','LoopGain_HO')
        else:
            self.LoopGain_HO = 0.5
            
        if self.check_config_key('RTC','SensorFrameRate_HO'):
            frameRate_HO = self.get_config_value('RTC','SensorFrameRate_HO')
        else:
            frameRate_HO = 500.0
            
        if self.check_config_key('RTC','LoopDelaySteps_HO'):
            delay_HO = self.get_config_value('RTC','LoopDelaySteps_HO')
        else:
            delay_HO = 2
                     
        if self.check_config_key('RTC','LoopGain_LO'):
            temp = self.get_config_value('RTC','LoopGain_LO')
            if temp != 'optimize':
                self.LoopGain_LO = temp
            else:
                self.LoopGain_LO = None
        else:
            self.LoopGain_LO = None
            
        if self.check_config_key('RTC','SensorFrameRate_LO'):
            frameRate_LO = self.get_config_value('RTC','SensorFrameRate_LO')
        else:
            frameRate_LO = None
            
        if self.check_config_key('RTC','LoopDelaySteps_LO'):
            delay_LO = self.get_config_value('RTC','LoopDelaySteps_LO')
        else:
            delay_LO = None
            
        if self.check_config_key('RTC','ResidualError'):
            wfe = self.get_config_value('RTC','ResidualError')
        else:
            wfe = None
            
        self.rtc = rtc(self.LoopGain_HO, frameRate_HO, delay_HO, wfe=wfe,
                       loopGainLO=self.LoopGain_LO, frameRateLO=frameRate_LO, delayLO=delay_LO)
               
#%% DEFORMABLE MIRRORS
        if not(self.check_section_key('DM')):
            self.raiseMissingRequiredSec('DM')

        if self.check_config_key('DM','NumberActuators'):
            nActu = self.get_config_value('DM','NumberActuators')
        else:
            self.raiseMissingRequiredOpt('DM', 'NumberActuators')
        
        if self.check_config_key('DM','DmPitchs'):
            DmPitchs = np.array(self.get_config_value('DM','DmPitchs'))
        else:
            self.raiseMissingRequiredOpt('DM','DmPitchs')
            self.error = True
            return
        
        if self.check_config_key('DM','InfModel'):
            InfModel = self.get_config_value('DM','InfModel') 
        else:
            InfModel = 'gaussian'
            
        if self.check_config_key('DM','InfCoupling'):
            InfCoupling = self.get_config_value('DM','InfCoupling') 
        else:
            InfCoupling = [0.2]
            
        if self.check_config_key('DM','DmHeights'):
            DmHeights = self.get_config_value('DM','DmHeights') 
        else:
            DmHeights = [0.0]
            
        if self.check_config_key('DM','OptimizationWeight'):
            opt_w = self.get_config_value('DM','OptimizationWeight') 
        else:
            opt_w = [0.0]
            
        if self.check_config_key('DM','OptimizationAzimuth'):
            opt_az = self.get_config_value('DM','OptimizationAzimuth') 
        else:
            opt_az = [0.0]
            
        if self.check_config_key('DM','OptimizationZenith'):
            opt_zen = self.get_config_value('DM','OptimizationZenith') 
        else:
            opt_zen = [0.0]

         # ----- verification
        if (len(opt_zen) != len(opt_az)) or (len(opt_zen) != len(opt_w)):
            self.raiseNotSameLength('DM', ['OptimizationZenith','OptimizationAzimuth','OptimizationWeight'])
              
        if self.check_config_key('DM','OptimizationConditioning'):
            cond = self.get_config_value('DM','OptimizationConditioning') 
        else:
            cond = 100.0
            
        if self.check_config_key('DM','NumberReconstructedLayers'):
            nrec = self.get_config_value('DM','NumberReconstructedLayers') 
        else:
            nrec = 10
            
        if self.check_config_key('DM','AoArea'):
            AoArea = self.get_config_value('DM','AoArea') 
        else:
            AoArea = 'circle'
        
        # ----- creating the dm class
        self.dms = deformableMirror(nActu, DmPitchs,
                                    heights=DmHeights, mechCoupling=InfCoupling,
                                    modes=InfModel,
                                    opt_dir=[opt_zen,opt_az],
                                    opt_weights=opt_w,
                                    opt_cond=cond,n_rec = nrec,
                                    AoArea=AoArea)
      
        #%% SCIENCE DETECTOR
        if not(self.check_section_key('sensor_science')):
            self.raiseMissingRequiredSec('sensor_science')
        
        if self.check_config_key('sensor_science','Name'):
            camName = self.get_config_value('sensor_science','Name')
        else:
            camName = 'SCIENCE CAM'
            
        if self.check_config_key('sensor_science','PixelScale'):
            psInMas = self.get_config_value('sensor_science','PixelScale')
        else:
            self.raiseMissingRequiredOpt('sensor_science', 'PixelScale')
        
        if self.check_config_key('sensor_science','FieldOfView'):
            fov = self.get_config_value('sensor_science','FieldOfView')
        else:
            self.raiseMissingRequiredOpt('sensor_science', 'FieldOfView')
        
        if self.check_config_key('sensor_science','Binning'):
            Binning = self.get_config_value('sensor_science','Binning')
        else:
            Binning = 1
            
        if self.check_config_key('sensor_science','SpotFWHM'):
            spotFWHM = self.get_config_value('sensor_science','SpotFWHM')
        else:
            spotFWHM = [[0.0,0.0,0.0]]
            
        if self.check_config_key('sensor_science','SpectralBandwidth'):
            bw = self.get_config_value('sensor_science','SpectralBandwidth')
        else:
            bw = 0.0
        if self.check_config_key('sensor_science','Transmittance'):
            tr = self.get_config_value('sensor_science','Transmittance')
        else:
            tr = [1.0]
        if self.check_config_key('sensor_science','Dispersion'):
            disp = self.get_config_value('sensor_science','Dispersion')
        else:
            disp = [[0.0],[0.0]]
        
        if self.check_config_key('sensor_science','NumberPhotons'):
            nphSC = self.get_config_value('sensor_science','NumberPhotons')
        else:
            nphSC = np.inf
        
        if self.check_config_key('sensor_science','Saturation'):
            saturation = self.get_config_value('sensor_science','Saturation')
        else:
            saturation = np.inf
            
        if self.check_config_key('sensor_science','SigmaRON'):
            ron = self.get_config_value('sensor_science','SigmaRON')
        else:
            ron = 0.0
        
        if self.check_config_key('sensor_science','Gain'):
            self.detectorGainScience = self.get_config_value('sensor_science','Gain')
        else:
            self.detectorGainScience = 1
            
        if self.check_config_key('sensor_science','SkyBackground'):
            sky = self.get_config_value('sensor_science','SkyBackground')
        else:
            sky = 0.0
        
        if self.check_config_key('sensor_science','Dark'):
            dark = self.get_config_value('sensor_science','Dark')
        else:
            dark = 0.0
        
        if self.check_config_key('sensor_science','ExcessNoiseFactor'):
            excess = self.get_config_value('sensor_science','ExcessNoiseFactor')
        else:
            excess = 1.0
        
        self.cam = detector(psInMas, fov,
                            binning=Binning, spotFWHM=spotFWHM, saturation=saturation,
                            nph=nphSC, bandwidth=bw, transmittance=tr, dispersion=disp,
                            gain=self.detectorGainScience, ron=ron, sky=sky, dark=dark, excess=excess,
                            tag=camName)
        
        # %% AO mode
        self.aoMode = 'SCAO'
        
        if self.lgs:
            if self.lgs.nSrc > 1:
                if self.dms.nDMs > 1:
                    self.aoMode = 'MCAO'
                else:
                    if self.dms.nRecLayers >1:
                        self.aoMode = 'LTAO' 
                    else:
                        self.aoMode = 'GLAO'                  
            else:
                self.aoMode = 'SLAO'
    
        # %% ERROR BREAKDOWN
        if self.rtc.holoop['gain'] > 0:
            self.errorBreakdown()


    def configLO(self):
        
        if not self.check_section_key('sources_LO'):
        
            print('Warning: No information about the tip-tilt star can be retrieved')            
        else:
            if self.check_config_key('sources_LO','Wavelength'):
                self.wvlGsLO = np.unique(np.array(self.get_config_value('sources_LO','Wavelength')))
            else:
                self.raiseMissingRequiredOpt('sources_LO','Wavelength')

            self.zenithGsLO   = self.get_config_value('sources_LO','Zenith')
            self.azimuthGsLO  = self.get_config_value('sources_LO','Azimuth')
            # ----- verification
            if len(self.zenithGsLO) != len(self.azimuthGsLO):
                self.raiseNotSameLength('sources_LO', ['Zenith','Azimuth'])

            self.ngs = source(self.wvlGsLO,self.zenithGsLO,self.azimuthGsLO,tag="NGS",verbose=True)

    
    def configLOsensor(self):
            
        if self.check_config_key('sensor_LO','PixelScale'):
            psInMas = self.get_config_value('sensor_LO','PixelScale')
        else:
            self.raiseMissingRequiredOpt('sensor_LO', 'PixelScale')
            self.error = True
            return

        if self.check_config_key('sensor_LO','FieldOfView'):
            fov = self.get_config_value('sensor_LO','FieldOfView')
        else:
            self.raiseMissingRequiredOpt('sensor_LO', 'FieldOfView')
            self.error = True
            return

        if self.check_config_key('sensor_LO','Binning'):
            Binning = self.get_config_value('sensor_LO','Binning')
        else:
            Binning = 1

        if self.check_config_key('sensor_LO','SpotFWHM'):
            spotFWHM = self.get_config_value('sensor_LO','SpotFWHM')
        else:
            spotFWHM = [[0.0, 0.0]]

        if self.check_config_key('sensor_LO','NumberPhotons'):
            nphLO = self.get_config_value('sensor_LO','NumberPhotons')
        else:
            nphLO = np.inf

        if self.check_config_key('sensor_LO','Gain'):
            self.detectorGainLO = self.get_config_value('sensor_LO','Gain')
        else:
            self.detectorGainLO = 1
            
        if self.check_config_key('sensor_LO','SigmaRON'):
            ron = self.get_config_value('sensor_LO','SigmaRON')
        else:
            ron = 0.0

        if self.check_config_key('sensor_LO','SkyBackground'):
            sky = self.get_config_value('sensor_LO','SkyBackground')
        else:
            sky = 0.0

        if self.check_config_key('sensor_LO','Dark'):
            dark = self.get_config_value('sensor_LO','Dark')
        else:
            dark = 0.0

        if self.check_config_key('sensor_LO','ExcessNoiseFactor'):
            excess = self.get_config_value('sensor_LO','ExcessNoiseFactor')
        else:
            excess = 1.0

        if self.check_config_key('sensor_LO','SpectralBandwidth'):
            bw = self.get_config_value('sensor_LO','SpectralBandwidth')
        else:
            bw = 0.0
        if self.check_config_key('sensor_LO','Transmittance'):
            tr = self.get_config_value('sensor_LO','Transmittance')
        else:
            tr = [1.0]
        if self.check_config_key('sensor_LO','Dispersion'):
            disp = self.get_config_value('sensor_LO','Dispersion')
        else:
            disp = [[0.0], [0.0]]

        if self.check_config_key('sensor_LO','NumberLenslets'):
            nL = self.get_config_value('sensor_LO','NumberLenslets')
        else:
            nL = [1]

        if self.check_config_key('sensor_LO','NoiseVariance'):
            NoiseVar = self.get_config_value('sensor_LO','NoiseVariance')
        else:
            NoiseVar = [None]

        if self.check_config_key('sensor_LO','Algorithm'):
            algorithm = self.get_config_value('sensor_LO','Algorithm')
        else:
            algorithm = 'wcog'

        if self.check_config_key('sensor_LO','WindowRadiusWCoG'):
            temp = self.get_config_value('sensor_LO','WindowRadiusWCoG')
            if temp != 'optimize':
                wr = temp
            else:
                wr = 1.0
        else:
            wr = 1.0

        if self.check_config_key('sensor_LO','ThresholdWCoG = 0.0'):
            thr = self.get_config_value('sensor_LO','ThresholdWCoG = 0.0')
        else:
            thr = 0.0

        if self.check_config_key('sensor_LO','NewValueThrPix'):
            nv = self.get_config_value('sensor_LO','NewValueThrPix')
        else:
            nv = 0.0
            
        self.tts = sensor(psInMas, fov,
                          binning=Binning, spotFWHM=spotFWHM,
                          nph=nphLO, bandwidth=bw, transmittance=tr, dispersion=disp,
                          gain=self.detectorGainLO, ron=ron, sky=sky, dark=dark, excess=excess,
                          nL=nL, dsub=list(self.D/np.array(nL)),
                          wfstype='Shack-Hartmann', noiseVar=NoiseVar,
                          algorithm=algorithm, algo_param=[wr,thr,nv], tag="TT WFS")

    
    def __repr__(self):
        
        s = '\t\t\t\t________________________'+ self.aoMode + ' SYSTEM ________________________\n\n'
        s += self.src.__repr__() + '\n'
        if self.lgs:
            s += self.lgs.__repr__() + '\n'
        if self.ngs:
            s += self.ngs.__repr__() + '\n'
        s += self.atm.__repr__() + '\n'
        s+= self.tel.__repr__() + '\n'
        s+= self.wfs.__repr__() +'\n'
        if self.tts:
            s+= self.tts.__repr__() +'\n'
        
        s+= self.rtc.__repr__() +'\n'
        s+= self.dms.__repr__() +'\n'
        s+= self.cam.__repr__()

        return s
    
    def errorBreakdown(self):
        """
            Computing the AO wavefront error breakdown based on theoretical formula
        """
        
        rad2nm = lambda x:  np.sqrt(x) * self.atm.wvl*1e9/2/np.pi
        self.wfe = dict()
        
        Dr053 = (self.tel.D/self.atm.r0)**(5/3)
        dactur053 = (self.dms.pitch[0]/self.atm.r0)**(5/3)
        dsubr053 = (self.wfs.optics[0].dsub/self.atm.r0)**(5/3)
        
        # DM fitting error
        self.wfe['DM fitting'] = rad2nm(0.23*dactur053)
        
        # Aliasing error
        self.wfe['WFS aliasing'] = rad2nm(0.07*dsubr053)      
        
        # Servo-lag errors
        ff = np.pi*0.5**2 # to account for the loss of valid actuator outside the pupil
        nMax = int(np.sqrt(ff)*(self.dms.nControlledRadialOrder[0]+1))
        if hasattr(self.rtc,'ttloop') and self.tts !=None :
            nMin = 3
        else:
            nMin = 1
            
        nrad  = np.array(range(nMin,nMax))
        self.wfe['HO Servo-lag'] = rad2nm(0.04 * (self.atm.meanWind/self.tel.D/self.rtc.holoop['bandwidth'])\
                                    * Dr053 * np.sum((nrad+1)**(-2/3)))
        
        # Noise errors        
        if self.wfs.processing.noiseVar == [None]:
            varNoise = self.wfs.NoiseVariance(self.atm.r0 ,self.atm.wvl)
        else:
            varNoise = self.wfs.processing.noiseVar
            
        self.wfe['HO Noise'] = rad2nm(np.mean(varNoise))
        
        if hasattr(self.rtc,'ttloop') and self.tts !=None:
            self.wfe['TT Servo-lag'] = rad2nm(0.04 * (self.atm.meanWind/self.tel.D/self.rtc.ttloop['bandwidth'])* Dr053 * 2**(-2/3))
            
            if self.tts.processing.noiseVar == [None]:
                #varNoise = self.tts.NoiseVariance(self.atm.r0 ,self.atm.wvl) # this line has been commented because it makes no sense
                                                                              # to consider the open loop r0 in this computation:
                                                                              # we expect the LO WFS to use a spot corrected by HO loop
                varNoise = self.tts.NoiseVariance(self.tts.optics[0].dsub ,self.atm.wvl)
            else:
                varNoise = self.tts.processing.noiseVar
            
            self.wfe['TT Noise']     = rad2nm(np.mean(varNoise))
        else:
            self.wfe['TT Servo-lag'] = 0
            self.wfe['TT Noise'] = 0
        
        # Focal anisoplanatism
        if self.lgs and self.lgs.height[0] > 0:
            self.wfe['Focal anisoplanatism'] = anisoplanatismModel.focal_anisoplanatism_variance(self.tel,self.atm,self.lgs)
        else:
            self.wfe['Focal anisoplanatism'] = 0
            
        # TO be added : angular anisoplanatisms
           
        self.wfe['Total'] = np.sqrt(self.wfe['DM fitting']**2 + self.wfe['WFS aliasing']**2\
                                    + self.wfe['HO Servo-lag']**2 + self.wfe['HO Noise']**2\
                                    + self.wfe['TT Servo-lag']**2 + self.wfe['TT Noise']**2\
                                    + self.wfe['Focal anisoplanatism']**2)
        self.wfe['Strehl'] = np.exp(-self.wfe['Total']**2 * (2*np.pi*1e-9/self.src.wvl[0])**2)
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sun Mar  7 10:02:21 2021

@author: omartin
"""

#%% IMPORTING LIBRAIRIES
import numpy as np
import os
import matplotlib as mpl
import matplotlib.pyplot as plt
from distutils.spawn import find_executable
import matplotlib.ticker as mtick
import time
from astropy.table import QTable
from scipy.stats import linregress

import p3.aoSystem.FourierUtils as FourierUtils
from p3.psfao21.psfao21 import psfao21
from p3.psfFitting.psfFitting import psfFitting
np.random.seed(1)

#%% CLASS
class deepLoopPerformance:
    ''' neuralNetworkPerformance class that allows to assess the 
    reconstruction performance of DEEPLOOP from the output .txt files
    '''
    
    def __init__(self,path_txt,path_ini=None,path_save=None,path_root='',nPSF=1000,
                 fit=False,mag=0,zP=25.44,DIT=0.5,nDIT=50,skyMag=13.6,ron=60,init='truth'):
        '''
        path_txt can be either:
            - a string of characters that gives the path to the txt file
            - a tuple of string for multi-configurations analyses
        '''
        # PARSING INPUTS
        # txt files created by DEEPLOOP
        if type(path_txt) == str:
            path_txt = [path_txt]
        self.path_txt = np.unique(path_txt)
        self.nCases   = len(self.path_txt)
        # verifying the non-emptiness of the path_txt field
        if self.nCases == 0:
            print('WARNING : the field path_txt is empty; no analyses possible.')
            return
        # verifying the existence of the txt files
        for k in range(self.nCases):
            if not os.path.isfile(self.path_txt[k]):
                print('ERROR : the path #%d does not exist'%(k+1))
                return
            
        # .ini file : needed for PSF computation
        self.path_ini = path_ini
        if self.path_ini == None:
            print('WARNING : the field path_ini is None : no analyses on PSFs possible')
        
        # IDENTIFYING DATA
        self.idNN     = np.empty(self.nCases,dtype=list)
        self.idData   = np.empty(self.nCases,dtype=list)
        self.dataType = np.empty(self.nCases,dtype=list)
        self.nParam   = np.empty(self.nCases,dtype=list)
            
        for n in range(self.nCases):
            # data identification from the file name
            s = self.path_txt[n].split('/')[-1]
            s = s.split('_')
            self.idNN[n]     = s[1]
            self.idData[n]   = s[2]
            self.dataType[n] = s[0][3:]
            # reading the first line to count the number of parameters
            self.nParam[n]   = self.read_txt_files(self.path_txt[n],getParamNumberOnly=True)
            
        self.nNetworks = len(self.idNN)
        self.nDataSets = len(self.idData)
        print('Test of %d network architectures performance on %d data sets'%(self.nNetworks,self.nDataSets))
        
        # INSTANTIATING DATA STRUCTURES
        self.gtruth = np.empty(self.nCases,dtype=tuple)
        self.nnest  = np.empty(self.nCases,dtype=tuple)
        self.labels = np.empty(self.nCases,dtype=tuple)
        self.metrics_param = np.empty(self.nCases,dtype=tuple)
        for n in range(self.nCases):
            # get parameters
            self.gtruth[n],self.nnest[n],self.labels[n] = self.read_txt_files(self.path_txt[n])
            # get metrics
            self.metrics_param[n] = np.zeros((self.nParam[n],7))
            for k in range(self.nParam[n]):
                self.metrics_param[n][k] = self.get_parameters_metrics(self.gtruth[n][k],self.nnest[n][k])
                
            self.path_save = path_save
            if self.path_save:
                Tab = QTable(names=self.labels[n])   
                for k in range(7):
                    Tab.add_row(self.metrics_param[n][:,k])
                Tab.write(path_save + self.idNN[n] + '_' + self.idData[n] + '_' + self.dataType[n] + '.csv',overwrite=True)
            
        
        # COMPUTING PSF METRICS
        if self.path_ini:
            self.fit = fit
            self.psfao = psfao21(path_ini,path_root=path_root)
            #if type(init) == str:
            self.init_fit = init
            #else:
            #    self.init_fit = str(init)
            self.get_psf_metrics(nPSF=nPSF,fit=self.fit,init=self.init_fit,mag=mag,zP=zP,DIT=DIT,nDIT=nDIT,skyMag=skyMag,ron=ron)
        
    def __call__(self,fontsize=22,fontfamily='normal',fontserif='Palatino',
                 figsize=20,constrained_layout=True,nBins=100,nstd=np.inf):
        '''
        Display DEEPLOOP performance
        '''
        # managing display configuration
        
        plt.close('all')
        # format
        font = {'weight' : 'normal',
        'size'   : fontsize,
        'family': fontfamily,
        'serif': fontserif}
        mpl.rc('font', **font)
        # latex
        if find_executable('tex'): 
            usetex = True
        else:
            usetex = False
        text = {'usetex': usetex}
        mpl.rc('text',**text)

        
        #number of axes in scientific notation
        formatter = mtick.ScalarFormatter(useMathText=False)
        formatter.set_scientific(True) 

        # ------ VERSUS PLOTS
        for n in range(self.nCases):
            # creating the figure
            nP = self.nParam[n]
            k1 = int(np.sqrt(nP))
            k2 = int(nP/k1)
            fig , axs = plt.subplots(k1,k2,figsize=(int(figsize*k2),int(figsize*k1)),constrained_layout=constrained_layout)
            a=-1
            for m in range(nP):
                b = m%k2
                if b==0:
                    a+=1                    
                
                mn = min(self.gtruth[n][m].min(),self.nnest[n][m].min())
                mx = max(self.gtruth[n][m].max(),self.nnest[n][m].max())
                axs[a,b].plot(self.gtruth[n][m],self.nnest[n][m],'bo')
                axs[a,b].plot([mn,mx],[mn,mx],'k--')
                axs[a,b].set_xlabel(self.labels[n][m] + ' simulated',fontsize=fontsize)
                axs[a,b].set_ylabel(self.labels[n][m] + ' reconstructed',fontsize=fontsize)
                #axs[a,b].set_aspect('equal')
                if mx > 10 or mn < 0.1:
                    formatter.set_powerlimits((-1,1)) 
                    axs[a,b].yaxis.set_major_formatter(formatter) 
                    axs[a,b].xaxis.set_major_formatter(formatter) 
            
            # saving 
            if self.path_save:
                path_fig = self.path_save + 'versusplots_parameters_' + self.dataType[n]\
                            + 'data_' + self.idNN[n] + '_' + str(self.idData[n])\
                            + '_' + str(len(self.nnest[n][0])) 
                plt.savefig(path_fig)    
        
        # ------ HISTOGRAMS
        for n in range(self.nCases):
            # creating the figure
            nP = self.nParam[n]
            k1 = int(np.sqrt(nP))
            k2 = int(nP/k1)
            fig , axs = plt.subplots(k1,k2,figsize=(int(figsize*k2),int(figsize*k1)),constrained_layout=constrained_layout)
            a=-1
            for m in range(nP):
                b = m%k2
                if b==0:
                    a+=1 
                b = m%k2
                
                if m < 6:
                    err = 1e2*(self.nnest[n][m] - self.gtruth[n][m])/self.gtruth[n][m]
                    lab = '\%'
                else:
                    err = self.nnest[n][m] - self.gtruth[n][m]
                    lab = 'nm'
                    
                err = err[abs(err)<nstd*err.std()]
                axs[a,b].hist(err, weights=np.ones_like(err) / len(err), bins=nBins)
                axs[a,b].set_xlabel(self.labels[n][m] + ' error ('+lab+')')
                axs[a,b].set_ylabel('Probability')
                axs[a,b].plot([0,0],axs[a,b].get_ylim(),'k--')
                if nstd != np.inf:
                    axs[a,b].set_xlim([-nstd*err.std(),nstd*err.std()])
            
            # saving 
            if self.path_save:
                path_fig = self.path_save + 'histograms_parameters_' + self.dataType[n]\
                            + 'data_' + self.idNN[n] + '_' + str(self.idData[n])\
                            + '_' + str(len(self.nnest[n][0])) 
                plt.savefig(path_fig)   
                
        # ------ PSFs METRICS
        if self.path_ini:
            for n in range(self.nCases):
                # creating the figure
                fig , axs = plt.subplots(2,2,figsize=(figsize,figsize),constrained_layout=constrained_layout)
                # MSE                
                err = self.mse[n]
                nPSF= err.shape[1]
                if self.fit:      
                    err    = self.mse[n][0]
                    errfit = self.mse[n][1]
                    err    = err[err==err]
                    errfit = errfit[errfit==errfit]
                    err    = err[err <= nstd*err.std()]
                    errfit = errfit[errfit <= nstd*errfit.std()]
                    mn     = 0
                    mx     = max([max(err),max(errfit)])
                    bins   = np.linspace(mn,mx,num=int(min(nPSF/10,nBins)))           
                    axs[0,0].hist(err, weights=np.ones_like(err) / len(err), bins=bins,label='DEEPLOOP',alpha=0.5)
                    axs[0,0].hist(errfit, weights=np.ones_like(errfit) / len(errfit), bins=bins,label='PSF-FITTING',alpha=0.5)
                    axs[0,0].legend()
                else:
                    err = err[0][err[0]==err[0]]
                    axs[0,0].hist(err[0], weights=np.ones_like(err[0]) / len(err[0]), bins=int(min(nPSF/10,nBins)))
                
                axs[0,0].set_xlabel('Mean square error (\%)')
                axs[0,0].set_ylabel('Probability')
                #if nstd != np.inf:
                #    axs[0,0].set_xlim([0,nstd*err[0].std()])
                
                # SR
                if self.fit:
                    err    = 1e2*(self.SR[n][1] - self.SR[n][0])/self.SR[n][0]
                    errfit = 1e2*(self.SR[n][2] - self.SR[n][0])/self.SR[n][0]
                    err    = err[err==err]
                    errfit = errfit[errfit==errfit]
                    err    = err[abs(err) <= np.median(err) + nstd*err.std()]
                    errfit = errfit[abs(errfit) <= np.median(errfit) + nstd*errfit.std()]
                    mn     = min([min(err),min(errfit)])
                    mx     = max([max(err),max(errfit)])
                    bins   = np.linspace(mn,mx,num=int(min(nPSF/10,nBins)))           
                    axs[0,1].hist(err, weights=np.ones_like(err) / len(err), bins=bins,label='DEEPLOOP',alpha=0.5)
                    axs[0,1].hist(errfit, weights=np.ones_like(errfit) / len(errfit), bins=bins,label='PSF-FITTING',alpha=0.5)
                    axs[0,1].legend()
                else:
                    err = 1e2*(self.SR[n][1] - self.SR[n][0])/self.SR[n][0]
                    err = err[err==err]
                    axs[0,1].hist(err, weights=np.ones_like(err) / len(err), bins=int(min(nPSF/10,nBins)))
                
                axs[0,1].set_xlabel('Strehl-ratio error (\%)')
                axs[0,1].set_ylabel('Probability')
                axs[0,1].plot([0,0],axs[0,1].get_ylim(),'k--')
                #if nstd != np.inf:
                #    axs[0,1].set_xlim([-nstd*err.std(),nstd*err.std()])
                
                # FWHM
                if self.fit:
                    err    = 1e2*(self.FWHM[n][1] - self.FWHM[n][0])/self.FWHM[n][0]
                    errfit = 1e2*(self.FWHM[n][2] - self.FWHM[n][0])/self.FWHM[n][0]
                    err = err[err==err]
                    errfit = errfit[errfit==errfit]
                    err    = err[abs(err) <= np.median(err) + nstd*err.std()]
                    errfit = errfit[abs(errfit) <= np.median(errfit) + nstd*errfit.std()]
                    mn     = min([min(err),min(errfit)])
                    mx     = max([max(err),max(errfit)])
                    bins   = np.linspace(mn,mx,num=int(min(nPSF/10,nBins)))
                    axs[1,0].hist(err, weights=np.ones_like(err) / len(err), bins=bins,label='DEEPLOOP',alpha=0.5) 
                    axs[1,0].hist(errfit, weights=np.ones_like(errfit) / len(errfit), bins=bins,label='PSF-FITTING',alpha=0.5)
                    axs[1,0].legend()
                else:
                    err = 1e2*(self.FWHM[n][1] - self.FWHM[n][0])/self.FWHM[n][0]
                    err = err[err==err]
                    axs[1,0].hist(err, weights=np.ones_like(err) / len(err), bins=int(min(nPSF/10,nBins)))
                
                axs[1,0].set_xlabel('FWHM error (\%)')
                axs[1,0].set_ylabel('Probability')
                axs[1,0].plot([0,0],axs[1,0].get_ylim(),'k--')
                #if nstd != np.inf:
                #    axs[1,0].set_xlim([-nstd*err.std(),nstd*err.std()])
                
                # Photometry
                if self.fit:
                    err    = self.mag_err[n][0]
                    errfit = self.mag_err[n][1]
                    err = err[err==err]
                    errfit = errfit[errfit==errfit]
                    err    = err[abs(err) <= np.median(err) + nstd*err.std()]
                    errfit = errfit[abs(errfit) <= np.median(errfit) + nstd*errfit.std()]
                    mn     = min([min(err),min(errfit)])
                    mx     = max([max(err),max(errfit)])
                    bins   = np.linspace(mn,mx,num=int(min(nPSF/10,nBins)))
                    axs[1,1].hist(err, weights=np.ones_like(err) / len(err), bins=bins,label='DEEPLOOP',alpha=0.5)
                    axs[1,1].hist(errfit, weights=np.ones_like(errfit) / len(errfit), bins=bins,label='PSF-FITTING',alpha=0.5)
                    axs[1,1].legend()
                else:
                    err = self.mag_err[n][0]
                    err = err[err==err]
                    axs[1,1].hist(err, weights=np.ones_like(err) / len(err), bins=int(min(nPSF/10,nBins)))
                
                axs[1,1].set_xlabel('Photometric error (mag)')
                axs[1,1].set_ylabel('Probability')
                axs[1,1].plot([0,0],axs[1,1].get_ylim(),'k--')
                #if nstd != np.inf:
                #    axs[1,1].set_xlim([-nstd*err.std(),nstd*err.std()])
                
                # saving 
                if self.path_save:
                    path_fig = self.path_save + 'histograms_psf_' + self.dataType[n]\
                                + 'data_' + self.idNN[n] + '_' + str(self.idData[n])\
                                + '_' + str(len(self.nnest[n][0])) + '_init_' + str(self.init_fit) +'.png'
                    plt.savefig(path_fig) 
                
                
            # ------ PSFs PLOTS
                plt.figure(figsize=(figsize,figsize),constrained_layout=constrained_layout)
                nPx = self.psf_mean[n].shape[0]
                fov = nPx * self.psfao.ao.cam.psInMas
                x = np.linspace(-fov/2,fov/2,num=nPx)
                plt.semilogy(x,self.psf_mean[n][nPx//2,:],'k',label='Mean PSF')
                plt.semilogy(x,self.psf_diff_mean[n][nPx//2,:],'b',label='Mean differential PSF')
                plt.semilogy(x,self.psf_diff_std[n][nPx//2,:],'r',label='Std differential PSF')
                if self.fit:
                    plt.semilogy(x,self.psf_diff_mean_fit[n][nPx//2,:],'g--',label='Mean differential PSF - fitting')
                    plt.semilogy(x,self.psf_diff_std_fit[n][nPx//2,:],'m--',label='Std differential PSF - fitting')
                plt.legend()
                plt.xlabel('Separation from the optical axis [mas]')
                # saving
                if self.path_save:
                    path_fig = self.path_save + 'psf_plots' + self.dataType[n]\
                                + 'data_' + self.idNN[n] + '_' + str(self.idData[n])\
                                + '_' + str(len(self.nnest[n][0])) + '_init_' + str(self.init_fit) +'.png'
                    plt.savefig(path_fig) 
                    
    def read_txt_files(self,path_txt,getParamNumberOnly=False):
        '''
        Reading the .txt input file and populating the gtruth and nnest arrays
        '''
        def is_float(string):
            try:
                float(string)
                return True
            except ValueError:
                return
            
        # GETTING THE NUMBER OF PARAMETERS
        tmp = open(path_txt)
        firstLine = tmp.readline().split()
        lab       = tmp.readline().split()
        labels    = lab[1:int((len(lab)-1)/2+1)]
        nParam = 0
        for n in range(len(firstLine)):
            if is_float(firstLine[n]):
                nParam+=1
        if getParamNumberOnly:
            tmp.close()
            return nParam
        
        # GETTING THE NUMBER OF LINES
        nData = sum(1 for line in tmp) -1
        tmp.close()
        
        # READING THE ENTIRE FILE
        groundTruth = np.zeros((nParam,nData))
        nnEstimates = np.zeros((nParam,nData))
        tmp = open(path_txt)
        lines = tmp.readlines()[2:]
        for n in range(nData):
            tmpP = np.array(lines[n][0:-1].split('\t')).astype(float)
            groundTruth[:,n] = tmpP[0:nParam]
            nnEstimates[:,n] = tmpP[nParam:]
            
        tmp.close()
        return groundTruth,nnEstimates,labels
            
    def get_parameters_metrics(self,gtruth,nnEst):
        '''
        getting median, mean, std and Peason coefficients
        '''
        
        def get_sub_metrics(vec):
            return np.mean(vec), np.median(vec), np.std(vec)
        
        # absolute difference
        dabs = nnEst - gtruth
        dabs_mean, dabs_med, dabs_std = get_sub_metrics(dabs)
        # relative difference
        drel = 1e2 * dabs/gtruth
        drel_mean, drel_med, drel_std = get_sub_metrics(drel)
        # Pearson
        coeff_pearson = np.corrcoef(gtruth,nnEst)

        return dabs_mean, dabs_med, dabs_std,drel_mean, drel_med, drel_std,coeff_pearson[0,1]
        
    def get_psf_metrics(self,nPSF=1000,fit=False,mag=0,init=0.1,zP=25.44,DIT=0.5,nDIT=50,skyMag=13.6,ron=60):
        
        t0 = time.time()
        print('Regenerating the PSFs ... ')
        p2x = lambda x: list(x[0:5]) + [0] + [x[5]] + [0,0,0,1,0,0,0] + list(x[6:-1])
        
        self.mag_err = np.empty(self.nCases,dtype=list)
        self.mse     = np.empty(self.nCases,dtype=list)
        self.SR      = np.empty(self.nCases,dtype=list)
        self.FWHM    = np.empty(self.nCases,dtype=list)
        
        self.psf_mean      = np.empty(self.nCases,dtype=list)
        self.psf_diff_mean = np.empty(self.nCases,dtype=list)
        self.psf_diff_std  = np.empty(self.nCases,dtype=list)
        if fit:
            self.psf_diff_mean_fit = np.empty(self.nCases,dtype=list)
            self.psf_diff_std_fit  = np.empty(self.nCases,dtype=list)
            self.fitEstimates      = np.empty(self.nCases,dtype=list) 
            p_median               = np.median(self.gtruth[0],axis=1)
        nPx = self.psfao.ao.cam.fovInPix
        nC  = 2 + fit*1
        
        for n in range(self.nCases):
            # instantiating outputs
            nPSFtot          = self.gtruth[n].shape[1]
            self.mag_err[n]  = np.zeros((nC-1,nPSF))
            self.mse[n]      = np.zeros((nC-1,nPSF))
            self.SR[n]       = np.zeros((nC,nPSF))
            self.FWHM[n]     = np.zeros((nC,nPSF))
            
            self.psf_mean[n] = np.zeros((nPx,nPx))
            self.psf_diff_mean[n] = np.zeros((nPx,nPx))
            self.psf_diff_std[n]  = np.zeros((nPx,nPx))
            if fit:
                self.fitEstimates[n]      = np.zeros((self.nParam[n],nPSF))
                self.psf_diff_mean_fit[n] = np.zeros((nPx,nPx))
                self.psf_diff_std_fit[n]  = np.zeros((nPx,nPx))  
                if type(init) == float:
                    frandn = np.random.normal(1,0.2,(nPSF,self.nParam[n]))
                    
            # down-selection 
            idx = np.random.randint(0,high=nPSFtot,size=nPSF)
            if self.nParam[n] == 6:
                nModes = 0
            else:
                nModes = self.psfao.ao.tel.nModes
                
            for k in range(nPSF):
                # GETTING THE GROUND TRUTH AND THE ESTIMATED PSF
                psf_true = np.squeeze(self.psfao( p2x(self.gtruth[n][:,idx[k]]) ))
                psf_ml   = np.squeeze(self.psfao( p2x(self.nnest[n][:,idx[k]]) )) 
                
                # PHOTOMETRY
                tmp,_,_,_,_     = linregress(psf_true.reshape(-1),psf_ml.reshape(-1))
                self.mag_err[n][0,k] = -2.5*np.log10(tmp)
                
                # AVERAGE PSF
                self.psf_mean[n]      += psf_true/nPSF
                self.psf_diff_mean[n] += abs(psf_ml - psf_true)/nPSF
                self.psf_diff_std[n]  += (psf_ml - psf_true)**2 /nPSF
                
                # GETTING METRICS
                self.mse[n][0,k]  = 1e2 * np.sqrt(np.sum((psf_ml-psf_true)**2))/psf_true.sum()
                self.SR[n][0,k]   = FourierUtils.getStrehl(psf_true,self.psfao.ao.tel.pupil,self.psfao.freq.sampRef) 
                self.SR[n][1,k]   = FourierUtils.getStrehl(psf_ml,self.psfao.ao.tel.pupil,self.psfao.freq.sampRef) 
                self.FWHM[n][0,k] = FourierUtils.getFWHM(psf_true,self.psfao.ao.cam.psInMas,nargout=1) 
                self.FWHM[n][1,k] = FourierUtils.getFWHM(psf_ml,self.psfao.ao.cam.psInMas,nargout=1) 
                
                # PSF-FITTING
                if fit:
                    weights = np.ones_like(psf_true)
                    if mag != 0:
                        # noising the image
                        Flux      = 10 ** (-0.4*(mag - zP))*DIT*nDIT
                        skyFlux   = 10 ** (-0.4*(skyMag - zP)) * nDIT * DIT * (self.psfao.ao.cam.psInMas/1e3) ** 2
                        ronStack  = ron * np.sqrt(nDIT)
                        noise_sky = np.random.poisson(skyFlux*np.ones_like(psf_true))
                        noise_dec = ronStack*np.random.randn(psf_true.shape[0],psf_true.shape[1])
                        noise_dec-= noise_dec.min()
                        im_noise  = np.random.poisson(Flux*psf_true) + noise_sky  + noise_dec
                        # computing the weights
                        weights   = 1.0/np.sqrt(ronStack**2 + psf_true)   
                    else:
                        im_noise = psf_true
                        
                    # defining the initial guess
                    if init == 'truth':
                        x0 = p2x(self.gtruth[n][:,idx[k]])
                    elif init == 'median':
                        x0 = p2x(p_median)
                    elif type(init) == float:
                        if init > 1:
                            init/=100
                        p_r= self.gtruth[n][:,idx[k]] * frandn[k]
                        p_r= np.maximum(p_r,self.gtruth[n].min(axis=1))
                        p_r= np.minimum(p_r,self.gtruth[n].max(axis=1))
                        x0 = p2x(p_r)
                    else:
                        raise ValueError("The init field must be 'truth', 'median' or a float")                    
                    # fit the image; the image musn't be normalized to properly account for the Poisson noise

                    if mag != 0:
                        x0[10]   = Flux
                        normType = 0
                    else:
                        normType = 1
                    fixed= (False,)*5 + (True,False,) + (True,)*3 + (False,)*4 + (False,)*nModes

                    res  = psfFitting(im_noise,self.psfao,x0,fixed=fixed,verbose=-1,weights=weights,normType=normType)
                    
                    # Get parameters
                    self.fitEstimates[n][:,k] = list(res.x[0:5]) + [res.x[6]] + list(res.x[14:])
                    
                    # averaged fitted-PSF
                    self.psf_diff_mean_fit[n] += abs(res.psf - psf_true)/nPSF
                    self.psf_diff_std_fit[n]  += (res.psf - psf_true)**2 /nPSF
                
                    # compare PSFs
                    tmp,_,_,_,_     = linregress(psf_true.reshape(-1),res.psf.reshape(-1))
                    self.mag_err[n][1,k]= -2.5*np.log10(tmp)
                    self.mse[n][1,k]    = 1e2 * np.sqrt(np.sum((res.psf-psf_true)**2))/psf_true.sum()
                    self.SR[n][2,k]     = FourierUtils.getStrehl(res.psf,self.psfao.ao.tel.pupil,self.psfao.freq.sampRef) 
                    self.FWHM[n][2,k]   = FourierUtils.getFWHM(res.psf,self.psfao.ao.cam.psInMas,nargout=1)
                    
            self.psf_diff_std[n] = np.sqrt(self.psf_diff_std[n])
            if fit:
                self.psf_diff_std_fit[n] = np.sqrt(self.psf_diff_std_fit[n])
                
            print('... Done in %.2f s ! '%(time.time() - t0))
            
# 35421.20 for 10k cases, init=0.1


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Mar  6 17:28:53 2021

@author: omartin, mdumont
"""

#%% IMPORTING LIBRAIRIES
import pathlib
import numpy as np
from astropy.io import fits

import p3.aoSystem as aoSystemMain
from p3.aoSystem.pupil import pupil
from p3.aoSystem.segment import segment
from p3.aoSystem.zernike import zernike
from p3.aoSystem.spiders import spiders
import matplotlib.pyplot as plt
import p3.aoSystem.FourierUtils as FourierUtils
path_ao = str(pathlib.Path(aoSystemMain.__file__).parent.absolute())

#%% CREATING ELT PETALLING MODES 
nPx = 50
path_txt = path_ao + '/_txtFile/ELT_segmentVertices.txt'
spiRef = spiders([0, 60, 120], 0.5, symetric=True, D=39) 
eltPup = pupil(segClass=segment(6, 1.42/2, nPx), segCoord=path_txt, D=39,
                cobs=0.2375, spiderClass=spiRef, getPetal=True, fill_gap=True)
modes = eltPup.computeModes([1,2,3], area="petal")
plt.figure();plt.imshow(modes[:,0].sum(axis=0))
plt.figure();plt.imshow(modes[:,1].sum(axis=0))
plt.figure();plt.imshow(modes[:,2].sum(axis=0))

#%% CREATING KECK SEGMENTDED MODES
nPx = 200
path_txt = path_ao + '/_txtFile/Keck_segmentVertices.txt'
spiRef = spiders([0,60,120], 0.0254, symetric=True, D=10.5) 
pupSeg = pupil(segClass=segment(6,0.9,nPx),segCoord=path_txt, spiderClass=spiRef)
nSeg  = pupSeg.nSegments
# CREATING ZERNIKE MODES
modNumbers = list(range(1,3))
zerMod_list = zernike(modNumbers, pupSeg.matrix.shape[0])
nMod = len(modNumbers)

# CREATING THE SEGMENTED MODE
mode_lr = np.zeros((nMod,nPx,nPx))
mode_hr = np.zeros((nMod,pupSeg.matrix.shape[0],pupSeg.matrix.shape[0]))

for i in range(nMod):
    for j in range(nSeg):
        pupSeg.reset()
        pupSeg.applyPhaseErrorSegment([j], [1], [1])
        segShape = np.angle(pupSeg.matrix)
        segShape[segShape > 0] = 1
        intersec = segShape * zerMod_list.modes[i]

        val = np.mean(intersec[(segShape > 0)])
        pupSeg.applyPhaseErrorSegment([j],[1],[val])
        
        mode_hr[i] += np.angle(pupSeg.matrix)
        mode_lr[i] += FourierUtils.interpolateSupport(np.angle(pupSeg.matrix),nPx,kind='linear')
        
        pupSeg.reset()
    plt.figure()
    plt.imshow( mode_hr[i] )

# #%% SAVING
# path_save = '/Users/dumontmaxime/Desktop/LAM_stage/Results'
# hdu = fits.PrimaryHDU(mode_lr)
# hdu.writeto(path_save + '/keck_piston_modes_'+str(nPx)+'px_all_lr.fits',overwrite=True)

# hdu = fits.PrimaryHDU(mode_hr)
# hdu.writeto(path_save + '/keck_piston_modes_'+str(pupSeg.matrix.shape[0])+'px_all_hr.fits',overwrite=True)




#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Apr  8 09:37:55 2021

@author: omartin
"""

import numpy as np
from astropy.io import fits
from query_eso_archive import query_simbad
from astropy.coordinates import SkyCoord
from astropy.time import Time
from astropy import units as u
from sphere.transmission import irdis_nd
import os
import shutil

#%% IMAGE PROCESSING
def RemoveCorruptedData(path_data,path_out):
    """
        Read all data folders and move the data with corrupted PSF to the outliers folder. Corruption is tracked by bad centering of the PSF.
    """
    
    # list the data sets
    listData   = os.listdir(path_data)
    nData      = len(listData)
    
    # loop on all folders
    list_corr = []
    for kData in range(nData): # data folder
        path_data_k = path_data + listData[kData] +'/'        
        if os.path.isfile(path_data_k + 'ird_convert_recenter_dc5-SPH_SPARTA_PSFDATA-psf_sparta_data.fits'):
            path_fits = path_data_k + 'ird_convert_recenter_dc5-IRD_SCIENCE_PSF_MASTER_CUBE-median_unsat.fits'
            im   = fits.getdata(path_fits) #first dim : left/right, 2nd dim : nacquisition
            if im.ndim == 3:
                im = im[0]
            elif im.ndim == 4:
                im = im[-1,0]
            if im[32,32] != im.max():
                list_corr.append(listData[kData])
        else:
            list_corr.append(listData[kData])
            
    print('Detection of %d corrupted files'%(len(list_corr)))
    for n in range(len(list_corr)):
        print(list_corr[n])
        shutil.move(path_data+list_corr[n], path_out)
        
def ProcessCleanData(path_data,nPix):
    """
        Process pre-cleaned SPHERE cubes of coadds to recenter the frame, stack them and normalize them
    """
    # read data
    im = fits.getdata(path_data)
    hdr= fits.getheader(path_data)
    # processing
    nF    = im.shape[0]
    out   = np.zeros((nF,nPix,nPix))
    for kf in range(nF):
        tmp     = im[kf]
        x0,y0   = np.unravel_index(np.argmax(tmp, axis=None), tmp.shape)
        out[kf] = tmp[x0-nPix//2:x0+nPix//2,y0-nPix//2:y0+nPix//2]
    # stacking 
    out = np.mean(out,axis=0)
    # normalizing
    DIT = hdr['ESO DET SEQ1 DIT']
    ND  = irdis_nd(hdr['HIERARCH ESO INS COMB IFLT'],['HIERARCH ESO INS4 FILT2 NAME'])
    out = out/DIT/ND
    return out

#%% READING THE HEADER
    
def GetWavelength(path_data):
    """
        Read the file IRD_SCIENCE_LAMBDA_INFO to extract the filter information
    """
    wvl = 1e-6*fits.getdata(path_data + '/ird_convert_recenter_dc5-IRD_SCIENCE_LAMBDA_INFO-lam.fits')
    bw  = wvl.max() - wvl.min()
    wvl = wvl.mean()
    
    return wvl, bw

def ReadSpartaData(path_data,which='last'):
    """
        Read the file SPH_SPARTA_PSFDATA-psf_sparta_data to get the atmospheric parameters.
        if the files exists, this is a [3, npsf+1,4] array
        ○    dim 1 : r0, wind, strehl ratio, seeing 
        ○    dim 2 : 0: npsf-1 : for each PSF file.   Last line = stat combining all frames of allfiles
        ○    dim 3 : min, max, avg
        reminder:  The SPARTA Strehl ratio is provided at 1.6 micron and the turbulence parameters (r0, seeing) at 500nm.
    """
    
    path_sparta = path_data + '/ird_convert_recenter_dc5-SPH_SPARTA_PSFDATA-psf_sparta_data.fits'
    if os.path.isfile(path_sparta):
        sparta = fits.getdata(path_sparta)
        # number of acquisitions during the observation
        nPSF          = sparta.shape[1]-1 
        # note : nPSF must == nFiles/2 with raw images or im.shape[0]//2 with processed images
                
        r0     = np.array(sparta[0,:,:])    
        vspeed = np.array(sparta[1,:,:])
        SR     = np.array(sparta[2,:,:])
        seeing = np.array(sparta[3,:,:])
        
        if which == 'last':
            r0     = r0[nPSF,:]
            vspeed = vspeed[nPSF,:]
            SR     = SR[nPSF,:]
            seeing = seeing[nPSF,:]
            
        return r0, vspeed, SR, seeing

def GetStarCoordinates(hdr):
    """
        Returns the name of the object as well as its RA/DEC coordinates
    """
    # get the star coordinates and name
    return hdr['ESO OBS NAME'], float(hdr['RA']), float(hdr['DEC'])
    
def GetStarMagnitudes(hdr):
    """
        Query simbad to get the star magnitudes.
    """
    
    OB_NAME, RA, DEC = GetStarCoordinates(hdr)

    # query simbad
    DICT_SIMBAD= query_simbad(Time(hdr['DATE-OBS']),SkyCoord(RA*u.degree,DEC*u.degree),name=OB_NAME)
    
    # get magnitudes
    VMAG = RMAG = GMAG = JMAG = HMAG = KMAG = -1
    if 'simbad_FLUX_V' in DICT_SIMBAD:
        VMAG = DICT_SIMBAD['simbad_FLUX_V']
    if 'simbad_FLUX_R' in DICT_SIMBAD:
        RMAG = DICT_SIMBAD['simbad_FLUX_R']
    if 'simbad_FLUX_G' in DICT_SIMBAD:
        GMAG = DICT_SIMBAD['simbad_FLUX_G']
    if 'simbad_FLUX_J' in DICT_SIMBAD:
        JMAG = DICT_SIMBAD['simbad_FLUX_J']
    if 'simbad_FLUX_H' in DICT_SIMBAD:
        HMAG = DICT_SIMBAD['simbad_FLUX_H']
    if 'simbad_FLUX_K' in DICT_SIMBAD:
        KMAG = DICT_SIMBAD['simbad_FLUX_K']
    
    return VMAG, RMAG, GMAG, JMAG, HMAG, KMAG

def GetDetectorConfig(hdr):
    """
        Read the header to get the detector configuration.
    """
    if 'PIXSCAL' in hdr:
        psInMas   = float(hdr['PIXSCAL'])
    else:
        psInMas = -1
    gain      = float(hdr['ESO DET CHIP1 GAIN'])
    ron       = float(hdr['ESO DET CHIP1 RON'])
    DIT       = hdr['ESO DET SEQ1 DIT']
    NDIT      = hdr['ESO DET NDIT']
    NDSKIP    = hdr['ESO DET NDSKIP']  
    
    return psInMas, gain, ron, DIT, NDIT, NDSKIP

def ReadStrehlValue(hdr):
    """
    Read the Strehl-ratio values in the header
    """
    if 'SRMEAN' in hdr:
        SRMEAN  = float(hdr['SRMEAN'])         
        SRMIN   = float(hdr['SRMIN'])                                
        SRMAX   = float(hdr['SRMAX'])
        return SRMEAN, SRMIN, SRMAX
    else:
        print('No Strehl-ratio values in the header')
        return -1, -1, -1
    
def GetDate(hdr):
    """
    Get the date from the header
    """
    DATE_OBS  = hdr['DATE-OBS']
    return DATE_OBS[0:4]+DATE_OBS[5:7]+DATE_OBS[8:10]+DATE_OBS[11:13]+DATE_OBS[14:16]+DATE_OBS[17:19]

def GetTelescopePointing(hdr):
    """
    Get the telescope pointing from the header.
    """
    TELAZ     = float(hdr['ESO TEL AZ'])
    TELALT    = float(hdr['ESO TEL ALT'])
    airmass   = 0.5*(float(hdr['ESO TEL AIRM END']) + float(hdr['ESO TEL AIRM START']))     
    
    return TELAZ, TELALT, airmass

def GetAmbiParameters(hdr):
    """
    Get atmosphere parameters from the header.
    """
    tau0      = float(hdr['ESO TEL AMBI TAU0'])
    wDir      = float(hdr['ESO TEL AMBI WINDDIR'])
    wSpeed    = float(hdr['ESO TEL AMBI WINDSP'])
    RHUM      = float(hdr['ESO TEL AMBI RHUM'])
    pressure  = 0.5*(float(hdr['ESO TEL AMBI PRES START']) + float(hdr['ESO TEL AMBI PRES END']))
    fwhm_linobs= float(hdr['ESO TEL IA FWHMLINOBS']) 
    # note: corrected active optics FWHM of the SH spots (arcsec) 
    #(along the line of sight, PSF-based algorithm as described in 
    #Martinez et al. 2011)
    
    return fwhm_linobs, tau0, wSpeed, wDir, RHUM, pressure

def GetTemperature(hdr):
    """
        Get values from temperature sensors.
    """
    
    # temperature sensors
    TEMP = np.zeros(60)
    TEMP[0] = float(hdr['ESO TEL AMBI TEMP'])
    TEMP[1] = float(hdr['ESO TEL TH M1 TEMP'])
    for t in range(1,6):
        TEMP[t+2]   = float(hdr['ESO INS1 TEMP10'+str(t)+' VAL'])       
    INS = 0
    for iii in range(4):
        if 'ESO INS'+str(iii+1)+' TEMP401 ID' in hdr:
            INS = int(iii+1)
    if INS>0:
        for t in range(3,53):
            if t<12 or t>51:
                dd = -273 # from Kelvin to degrees Celsius
            else:
                dd = 0
            TEMP[t+7]   = float(hdr['ESO INS'+str(INS)+' TEMP4'+str(t+3).zfill(2)+' VAL']) + dd    

    return TEMP

def GetMoonPosition(hdr):
    """
        Returns the RA/DEC of the moon during the observation.
    """
    return hdr['ESO TEL MOON RA'], hdr['ESO TEL MOON DEC']

        
        



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Aug  2 18:48:47 2021

@author: omartin
"""

#%%
import numpy as np
import os
from scipy.io import loadmat
import warnings
warnings.simplefilter(action='ignore', category=FutureWarning)
from scipy.sparse import csc_matrix

#%%
class structtype():
    pass

class telemetryKASP:
    
    def __init__(self,path_mat,path_save='./'):
        
        # Check the telemetry path
        self.path_mat = path_mat
        if path_mat == None or path_mat == [] or path_mat == '':
            print('No telemetry file')
            self.path_mat = None
        else:
            # Check the telemetry file
            if os.path.isfile(path_mat) == False:
                raise ValueError('Error : the .mat file does not exist\n')
                return
        self.path_save = path_save
        
        # instantiating the fields
        self.instantiating_fields()
        
        # reading the mat file
        self.data_struct = loadmat(path_mat,struct_as_record=True)['data_struct']
        
        # populating fields
        self.populating_fields()
        
        
    def instantiating_fields(self):
        """
            Instantiating the data structures for the Keck AO system"
        """
        #NOTE : SO FAR I"M USING STRUCTURES, BUT WE COULD RELY ON CLASSES IN THE aoSystem FOLDER
        # telescope
        self.tel = structtype()
        self.tel.D           = None  # DM actuators pitch multiplied by number of actuators
        self.tel.cobs        = None # central obstruction ratio
        self.tel.resolution  = None   
        self.tel.pupilAngle  = 0
        self.tel.path_pupil  = ''
        self.tel.path_telstat= ''
        
        # atmosphere
        self.atm = structtype()
        self.atm.seeing_dimm = -1
        self.atm.seeing_mass = -1
        self.atm.Cn2_mass    = -1
        # wfs
        self.src = structtype()
        # NGS
        self.ngs = structtype()
        self.ngs.zenith = [0.0]
        self.ngs.azimuth = [0.0]
        # LGS
        self.lgs = structtype()
        self.lgs.zenith = [0.0]
        self.lgs.azimuth = [0.0]
        self.lgs.height = 90e3
    
        # wfs
        self.wfs             = structtype()  
        self.wfs.pixel_scale = None
        self.wfs.fov         = None
        self.wfs.ron         = 0.0
        self.wfs.nSl         = [None] # Number of slopes measurements within the pupil (x and y)
        self.wfs.nSubap      = [None]
        self.wfs.dsub        = [None]
        self.wfs.algo        = 'cog'
        
        self.wfs.type        = 'Shack-Hartmann'
        self.wfs.modulation  = None
        self.wfs.binning     = 1
        self.wfs.nph         = None
        self.wfs.spot_fwhm   = [[0.0,0.0,0.0]]
        self.wfs.bw          = 0.0
        self.wfs.tr          = [1.0]
        self.wfs.disp        = [[0.0],[0.0]]
        self.wfs.dark        = 0.0
        self.wfs.sky         = 0.0
        self.wfs.gain        = 1.0
        self.wfs.excess      = 1.0
        self.wfs.noiseVar    = [None]
        self.wfs.win         = 2.0
        self.wfs.thres       = 0.0
        self.wfs.new         = 0.0
        self.wfs.theta       = [0]
        
        # tipTilt
        self.tipTilt             = structtype()
        self.tipTilt.pixel_scale = None
        self.tipTilt.fov         = None
        self.tipTilt.ron         = 0.0
        self.tipTilt.nSubap      = [1]
        self.tipTilt.dsub        = [None]
        self.tipTilt.algo        = 'cog'
        self.tipTilt.tilt2meter  = 1
        self.tipTilt.binning     = 1
        self.tipTilt.nph         = None
        self.tipTilt.spot_fwhm   = [[0.0,0.0,0.0]]
        self.tipTilt.bw          = 0.0
        self.tipTilt.tr          = [1.0]
        self.tipTilt.disp        = [[0.0],[0.0]]
        self.tipTilt.dark        = 0.0
        self.tipTilt.sky         = 0.0
        self.tipTilt.gain        = 1.0
        self.tipTilt.excess      = 1.0
        self.tipTilt.nSubap      = [1]
        self.tipTilt.noiseVar    = [None]
        self.tipTilt.win         = 2.0
        self.tipTilt.thres       = 0.0
        self.tipTilt.new         = 0.0
        
        # dm
        self.dm              = structtype()
        self.dm.nActuators   = [None]   # 1D Number of actuators                                            
        self.dm.nCom         = [None]     # Number of total actuators within the pupil
        self.dm.pitch        = [None]
        self.dm.mechCoupling = [None]
        self.dm.heights      = [0.0]
        self.dm.modes        = ''        
        self.dm.opt_cond     = 1e2
        self.dm.opt_zen      = [0.0]
        self.dm.opt_azi      = [0.0]
        self.dm.opt_weight   = [1.0]
        self.dm.nrec         = 7
        self.dm.area         = 'square'
        
        # cam
        self.cam = structtype()
        self.cam.spot_fwhm   = [[0.0,0.0,0.0]]
        self.cam.bw          = 0.0
        self.cam.transmission= [1.0]
        self.cam.dispersion  = [[0.0],[0.0]]
        self.cam.ron         = 0.0
        self.cam.gain        = 1.0
        
        # rec
        self.rec = structtype()
        
        # mat
        self.mat = structtype()
        
        # holoop
        self.holoop = structtype()
        self.holoop.tf = structtype()
        # ttloop
        self.ttloop = structtype()
        self.ttloop.tf = structtype()
        
    def populating_fields(self):
        '''
            Populating fields of the telemetryKASP object from the .mat file
        '''
        rad2arc = 3600 * 180/np.pi
        
        # TELESCOPE
        self.obsdate          = self.data_struct['simu_date'][0,0][0]
        self.acqtime          = self.data_struct['simu_time'][0,0][0]
        self.tel.name         = 'OOMAO'
        self.tel.D            = float(self.data_struct['D'])
        self.tel.cobs         = float(self.data_struct['cobs'])
        self.tel.resolution   = int(self.data_struct['resolution'])
        self.tel.zenith_angle = float(self.data_struct['zenith_angle'])
        self.tel.airmass      = float(self.data_struct['airmass'])
        self.tel.pupil        = np.array(self.data_struct['pupil'][0,0])
        self.tel.map_ncpa     = np.array(self.data_struct['map_ncpa'][0,0])
        
        # ATMOSPHERE
        self.atm.wvl        = float(self.data_struct['wvl_atm'])
        self.atm.r0         = float(self.data_struct['r0'])
        self.atm.seeing     = 3600 * 180/np.pi * self.atm.wvl/self.atm.r0
        self.atm.L0         = float(self.data_struct['L0'])
        self.atm.Cn2Heights = list(self.data_struct['cn2_alt'][0,0][0])
        self.atm.Cn2Weights = np.array(self.data_struct['cn2_weight'][0,0][0])
        self.atm.Cn2        = list(self.atm.Cn2Weights * self.atm.r0**(-5/3))
        self.atm.Cn2Weights = list(self.atm.Cn2Weights)
        self.atm.wSpeed     = list(self.data_struct['wind_speed'][0,0][0])
        self.atm.wDir       = list(self.data_struct['wind_dir'][0,0][0])
        
        # CAMERA
        self.cam.name       = 'OOMAO'
        self.cam.wvl        = [float(self.data_struct['src_wvl'])]
        self.cam.zenith     = [float(self.data_struct['src_zenith'])*rad2arc]
        self.cam.azimuth    = [float(self.data_struct['src_azimuth'])]
        self.cam.image      = np.array(self.data_struct['psf'][0,0])
        self.cam.strehl     = float(self.data_struct['strehl'][0,0])
        self.cam.psInMas    = float(self.data_struct['cam_psInMas'])
        self.cam.fov        = int(self.data_struct['cam_fov'])*2
        #note : the factor 2 allows to avoid aliasing effect
        self.cam.ittime     = float(self.data_struct['cam_ittime'])
        self.cam.ron        = float(self.data_struct['cam_ron'])
        self.cam.path_ncpa  = ''
        
        # SOURCES
        self.ngs.wvl        = float(self.data_struct['ngs_wvl'])
        self.ngs.zenith     = [float(self.data_struct['ngs_zenith'])*rad2arc]
        self.ngs.azimuth    = [float(self.data_struct['ngs_azimuth'])]
        
        if len(self.data_struct['lgs_azimuth'][0,0]) > 0:
            self.lgs.wvl        = float(self.data_struct['lgs_wvl'])
            self.lgs.zenith     = [float(self.data_struct['lgs_zenith'])*rad2arc]
            self.lgs.azimuth    = [float(self.data_struct['lgs_azimuth'])]
            self.lgs.height     = float(self.data_struct['lgs_height'])
            self.wfs.wvl        = self.lgs.wvl
            self.aoMode         = 'LGS'
        else:
            self.lgs.wvl        = None
            self.lgs.zenith     = None
            self.lgs.azimuth    = None
            self.wfs.wvl        = self.ngs.wvl
            self.aoMode         = 'NGS'
            
        # SYSTEM MATRICES
        mech2opt = 2
        self.mat.M          = mech2opt*np.array(self.data_struct['mat_rec'][0,0]) 
        self.mat.DMTTRem    = np.array(self.data_struct['mat_dmttrem'][0,0]) 
        self.mat.SlopeTTRem = np.array(self.data_struct['mat_slttrem'][0,0]) 
        self.mat.TT2COM     = np.array(self.data_struct['mat_tt2com'][0,0]) 
        self.mat.SL2TT      = np.array(self.data_struct['mat_sl2tt'][0,0]) 
        self.mat.COM2TT     = np.array(self.data_struct['mat_com2tt'][0,0]) 
        self.mat.dmIF       = csc_matrix.toarray(self.data_struct['mat_dmif'][0,0])
        self.mat.Hdm        = np.array(self.data_struct['mat_dmfilter'][0,0])
        
        # HO WFS
        self.wfs.pixel_scale = float(self.data_struct['wfs_psInMas'])
        self.wfs.fov         = int(self.data_struct['wfs_fov'])
        self.wfs.ron         = float(self.data_struct['wfs_ron'])
        self.wfs.nSubap      = [int(self.data_struct['wfs_nsubap'])]
        self.wfs.dsub        = [float(self.data_struct['wfs_dsub'])]
        self.wfs.nSlopes     = int(self.data_struct['wfs_nslopes'])
        self.wfs.nExp        = int(self.data_struct['wfs_nexp'][0])
        self.wfs.slopes      = np.squeeze(self.data_struct['wfs_slopes'][0,0]).T
        self.wfs.slopes     -= np.mean(self.wfs.slopes,axis=0)
        
        # DM
        # note : the factor 2 allows to convert the mechanical OPD to the optical OPD
        self.dm.pitch           = [float(self.data_struct['dm_pitch'])]
        self.dm.com             = mech2opt*np.array(np.dot(self.mat.DMTTRem,np.squeeze(self.data_struct['dm_com'][0,0] ))).T
        self.dm.com            -= np.mean(self.dm.com,axis=0)
        self.dm.validActuators  = self.data_struct['dm_validactu'][0,0]
        self.dm.nActuators      = [int(self.data_struct['dm_nactu'])]
        self.dm.nCom            = int(self.data_struct['dm_ncom'])
        self.dm.mechCoupling    = [float(self.data_struct['dm_coupling'])]
        self.dm.heights         = [float(self.data_struct['dm_height'])]
        
        # RESIDUAL WAVEFRONT
        self.mat.R              = np.array(np.dot(self.mat.DMTTRem,np.dot(self.mat.M,self.mat.SlopeTTRem))).T
        self.rec.res            = np.dot(self.wfs.slopes,self.mat.R)
        self.rec.res           -= np.mean(self.rec.res,axis=0)
        self.rec.wfe            = 1e9*np.sqrt(np.sum(np.std(self.rec.res,axis=0)**2)/self.dm.nCom)
        
        # TIP-TILT
        if len(self.data_struct['lgs_azimuth'][0,0]) > 0:
            self.tipTilt.wvl         = float(self.data_struct['ngs_wvl'])
            self.tipTilt.pixel_scale = float(self.data_struct['tt_psInMas'])
            self.tipTilt.fov         = int(self.data_struct['tt_fov'])
            self.tipTilt.ron         = float(self.data_struct['tt_ron'])
            self.tipTilt.nExp        = int(self.data_struct['tt_nexp'][0])
            self.wfs.slopes          = np.dot(self.mat.SlopeTTRem,self.wfs.slopes.T).T
            self.tipTilt.tilt2meter  = self.tipTilt.pixel_scale * self.tel.D/1e3/206264.8
            self.tipTilt.slopes      = self.tipTilt.tilt2meter*np.array(np.squeeze(self.data_struct['tt_slopes'][0,0] )).T
        else:
            self.tipTilt.nExp        = self.wfs.nExp
            self.tipTilt.tilt2meter  = self.wfs.pixel_scale * self.tel.D/1e3/206264.8
            self.tipTilt.slopes      = np.dot(self.mat.COM2TT,np.dot(self.mat.M,self.wfs.slopes.T)).T
            self.wfs.slopes          = np.dot(self.mat.SlopeTTRem,self.wfs.slopes.T).T
        
        self.tipTilt.com = np.array(np.squeeze(self.data_struct['tt_com'][0,0] )).T
        self.tipTilt.com-= np.mean(self.tipTilt.com,axis=0)
        
        # TRANSFER FUNCTION
        self.holoop.lat      = float(self.data_struct['lat_ho'])
        self.holoop.freq     = float(self.data_struct['rate_ho'])
        self.holoop.gain     = float(self.data_struct['gain_ho'])
        self.holoop.tf.num   = [self.holoop.gain,0,0,0]
        self.holoop.tf.den   = [-1,0,0]     
        self.ttloop.lat      = float(self.data_struct['lat_tt'])
        self.ttloop.freq     = float(self.data_struct['rate_tt'])
        self.ttloop.gain     = float(self.data_struct['gain_tt'])
        self.ttloop.tf.num   = [self.ttloop.gain,0,0,0]
        self.ttloop.tf.den   = [-1,0,0]
        
        
        

        


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Sat Apr 17 13:57:21 2021

@author: omartin
"""

import numpy as np
from p3.aoSystem.optics import optics
from p3.aoSystem.detector import detector
from p3.aoSystem.processing import processing

class sensor:
    """
    Wavefront sensor class. This class is instantiated through three sub-classes:\
    optics, detector and processing.
    """
    
    def __init__(self,pixel_scale,fov,binning=1,spotFWHM=[0.0,0.0],\
                 nph=np.inf,bandwidth=0.0,transmittance=[1.0],dispersion=[[0.0],[0.0]],\
                 ron=0.0,gain=1.0,dark=0.0,sky=0.0,excess=1.0, \
                 wfstype='Shack-Hartmann',nL=[1],dsub=[1],nSides=4,modulation=None,\
                 algorithm='wcog', algo_param=[5,0,0], noiseVar=[None], tag='WAVEFRONT SENSOR'):
        
                
        self.wfstype = wfstype
        self.nWfs = len(nL)
        self.tag  = tag
        
        # optics class
        self.optics = [optics(nL=nL[k],dsub=dsub[k],nSides=nSides,wfstype=wfstype,modulation=modulation) for k in range(self.nWfs)]
        
        # detector class
        self.detector = detector(pixel_scale,fov,binning=binning,spotFWHM=spotFWHM,\
                 nph=nph,bandwidth=bandwidth,transmittance=transmittance,dispersion=dispersion,\
                 ron=ron,gain=gain,dark=dark,sky=sky,excess=excess)
        
        # processing class
        self.processing = processing(algorithm=algorithm,settings=algo_param,noiseVar=noiseVar)
    
    def __repr__(self):
        
        if self.nWfs == 1:
            s = '__ ' + self.tag + ' __\n' + '--------------------------------------------- \n'
        else:
            s = '__ '+str(self.nWfs) + ' WAVEFRONT SENSORS __\n' + '--------------------------------------------- \n'
        s += self.optics[0].__repr__() + '\n'
        s+= self.detector.__repr__() + '\n'
        s+= self.processing.__repr__() +'\n'
        s+= '---------------------------------------------\n'
        return s
        
    def NoiseVariance(self,r0,wvl):
        
        rad2arcsec = 3600 * 180 / np.pi

        # parsing inputs
        varNoise    = np.zeros(self.nWfs)
        nph         = np.array(self.detector.nph)
        pixelScale  = self.detector.psInMas/1e3 # in arcsec
        ron         = self.detector.ron
        for k in range(self.nWfs):
            # pixel per subaperture, N_s in Thomas et al. 2006
            # nPix**2 is the total number of pixels used in the CoG calculation
            nPix = self.detector.fovInPix
            dsub = self.optics[k].dsub
            
            if self.wfstype.upper() == 'SHACK-HARTMANN':
                # spot FWHM in pixels and without turbulence, N_samp in Thomas et al. 2006
                # The condition nD = 2 corresponds to the Nyquist sampling of the spots
                nD = max(1,rad2arcsec * wvl/dsub /pixelScale) 

                # The full width at half-maximum (FWHM) of the spot, N_T in Thomas et al. 2006
                # For diffraction-limited spots nT = nD = 2
                nT = max(1,np.hypot(max(self.detector.spotFWHM[0][0:2])/1e3,rad2arcsec*wvl/r0)/pixelScale)

                # for WCoG, Nw is the weighting function FWHM in pixel
                nW = self.processing.settings[0]
                if nW < nT:
                    print('WARNING: weighting function FWHM of the WCoG is smaller than the FWHM of the spot, forcing it to have the same size.')
                    nW = nT
              
                # tCoG parameters
                th = self.processing.settings[1]
                new_val_th = self.processing.settings[2]
                
                # read-out noise calculation & photo-noise calculation
                # from Thomas et al. 2006
                if self.processing.algorithm == 'cog':
                    varRON  = np.pi**2/3 * (ron**2 /nph[k]**2) * (nPix**2/nD)**2
                    varShot  = np.pi**2/(2*np.log(2)*nph[k]) * (nT/nD)**2
                if self.processing.algorithm == 'tcog':
                    # Here we consider that the pixel used in the computation
                    # are the ones where the PSF is above the 0.5 w.r.t. the maximum value,
                    # so, nPix**2 is subsituted by np.ceil(nT**2*np.pi/4)
                    varRON  = np.pi**2/3 * (ron**2 /nph[k]**2) * (np.ceil(nT**2*np.pi/4)/nD)**2
                    varShot  = np.pi**2/(2*np.log(2)*nph[k]) * (nT/nD)**2
                if self.processing.algorithm == 'wcog':
                    varRON  = np.pi**3/(32*np.log(2)**2) * (ron**2 /nph[k]**2) * (nT**2+nW**2)**4/(nD**2*nW**4)
                    varShot  = np.pi**2/(2*np.log(2)*nph[k]) * (nT/nD)**2 * (nT**2+nW**2)**4/((2*nT**2+nW**2)**2*nW**4)
                if self.processing.algorithm == 'qc':
                    if nT > nD:
                        k = np.sqrt(2*np.pi) * (nT/(2*np.sqrt(2*np.log(2))) / nD)
                    else:
                        k = 1
                    varRON  = k *  4*np.pi**2 * (ron/nph[k])**2
                    varShot  = k * np.pi**2/nph[k]
                    
                if varRON.any() > 3:
                    print('The read-out noise variance is very high (%.1f >3 rd^2), there is certainly smth wrong with your inputs, set to 0'%(varRON))
                    varRON = 0
                    
                if varShot.any() > 3:
                    print('The shot noise variance is very high (%.1f >3 rd^2), there is certainly smth wrong with your inputs, set to 0'%(varShot))
                    varShot = 0
                    
            if self.wfstype.upper() == 'PYRAMID':
                varRON  = 4*ron**2/np.mean(nph)**2
                varShot = nph[k]/np.mean(nph)**2

            varNoise[k] = varRON + self.detector.excess * varShot
        
        return varNoise



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug 16 16:34:02 2018

@author: omartin
"""
import numpy as nnp
from . import gpuEnabled, cp, np, nnp, rotate

from astropy.io import fits
import os.path as ospath
import re
import p3.aoSystem.FourierUtils as FourierUtils

class Attribute(object):
    pass

class telescope:
    """ Telescope class that defines the telescope characteristics.
    Inputs are:
        - D: telescope diameter
        - elevation: telescope elevation in degree
        - obsRatio: central obstruction ratio from 0 to 1
        - resolution: pupil pixels resolution
    """
    
    # DEPENDANT VARIABLES DEFINITION   
    def get_pupil(self):
        return self.p_pupil
                 
    def set_pupil(self,val):
        self.p_pupil = val
        self.resolution = val.shape[0]
        
    pupil = property(get_pupil,set_pupil)        
        
    @property
    def R(self):
        """Telescope radius in meter"""
        return self.D/2
    @property
    def area(self):
        """Telescope area in meter square"""
        return np.pi*self.R**2*(1-self.obsRatio**2)
    @property
    def pupilLogical(self):
        """Telescope area in meter square"""
        return self.pupil.astype(bool)
    @property
    def airmass(self):
        return 1/nnp.cos(self.zenith_angle*np.pi/180)
    
    # CONSTRUCTOR
    def __init__(self,D,resolution, zenith_angle=0.0,obsRatio=0.0,pupilAngle = 0.0,\
                 path_pupil='', path_static_on='', path_static_off='', path_static_pos='',\
                 path_apodizer='', path_statModes='', extraErrorNm=0.0, extraErrorExp=-2, \
                 extraErrorMin=0.0, extraErrorMax=0.0, extraErrorLoNm=0.0, extraErrorLoExp=-2, \
                 extraErrorLoMin=0.0, extraErrorLoMax=0.0, verbose=True):
        
        # PARSING INPUTS
        self.D                 = D            # primary mirror diameter in meters
        self.zenith_angle      = zenith_angle # telescope zenith angle in degrees
        self.obsRatio          = obsRatio     # secondary mirror diameter in D units
        self.resolution        = resolution   # pupil resolution in pixels
        self.verbose           = verbose
        self.pupilAngle        = pupilAngle
        self.path_pupil        = path_pupil
        self.path_static_on    = path_static_on
        self.path_static_pos   = path_static_off
        self.path_apodizer     = path_apodizer
        self.path_statModes    = path_statModes
        self.extraErrorNm      = extraErrorNm
        self.extraErrorExp     = extraErrorExp
        self.extraErrorMin     = extraErrorMin
        self.extraErrorMax     = extraErrorMax
        self.extraErrorLoNm    = extraErrorLoNm
        self.extraErrorLoExp   = extraErrorLoExp
        self.extraErrorLoMin   = extraErrorLoMin
        self.extraErrorLoMax   = extraErrorLoMax
        
        #----- PUPIL DEFINITION        
        
        pupil = [] 
        if path_pupil!= '' and ospath.isfile(path_pupil) == True and re.search(".fits",path_pupil)!=None:
            self.verb = True
            pupil = np.asarray(fits.getdata(path_pupil))
            pupil[pupil!=pupil] = 0
            if self.pupilAngle !=0.0:
                pupil = rotate(pupil,self.pupilAngle,reshape=False)
            if pupil.shape[0] != resolution:
                pupil = FourierUtils.interpolateSupport(pupil,resolution,kind='linear')
            self.pupil = pupil
        else:
            # build an annular pupil model
            th  = self.pupilAngle*nnp.pi/180
            x   = nnp.linspace(-D/2,D/2,resolution)
            X,Y = nnp.meshgrid(x,x)
            Xr  = X*nnp.cos(th) + Y*nnp.sin(th)
            Yr  = Y*nnp.cos(th) - X*nnp.sin(th)
            R   = nnp.hypot(Xr,Yr)
            P   = (R <= self.R) * (R > self.R*self.obsRatio)
            self.pupil = np.asarray(P)
            self.verb = False
            self.path_pupil= ''
    
    
        #----- NCPA
        
        if path_static_on != None and ospath.isfile(path_static_on) == True and re.search(".fits",path_static_on)!=None:
            self.opdMap_on = np.asarray(fits.getdata(path_static_on))
            self.opdMap_on[self.opdMap_on!=self.opdMap_on] = 0
            if self.pupilAngle !=0.0:
                self.opdMap_on = rotate(self.opdMap_on,self.pupilAngle,reshape=False)
            self.opdMap_on = FourierUtils.interpolateSupport(self.opdMap_on,resolution,kind='linear')
        else:
            self.opdMap_on = None
            self.path_static_on= ''
            
        #----- FIELD-DEPENDANT STATIC ABERRATIONS
        if path_static_off != None and ospath.isfile(path_static_off) == True and re.search(".fits",path_static_off)!=None:   
            if path_static_pos != None and ospath.isfile(path_static_pos) == True and re.search(".fits",path_static_pos)!=None:   
                opdMap_off = np.asarray(fits.getdata(path_static_off))
                opdMap_pos = np.asarray(fits.getdata(path_static_pos))
                if opdMap_pos.shape[0] != opdMap_off[0]:
                    raise ValueError('You must provide as many positions values as maps')
                else:
                    self.opdMap_off = opdMap_off
                    self.opdMap_pos = opdMap_pos
            else:
                raise ValueError('Positions (zenith in arcsec, azimuth in radian) of the field-dependent aberrations must be provided as well as the maps')
        else:
            self.opdMap_off = None #center of the fov
            self.opdMap_pos = None #center of the fov
            self.path_static_off= ''
            self.path_static_pos= ''
            
        #----- APODIZER
        if path_apodizer!= '' and ospath.isfile(path_apodizer) and re.search(".fits",path_apodizer)!=None:
            self.apodizer = np.asarray(fits.getdata(path_apodizer))
            self.apodizer[self.apodizer!=self.apodizer] = 0
            if self.pupilAngle !=0.0:
                self.apodizer = rotate(self.apodizer,self.pupilAngle,reshape=False)
            self.apodizer = FourierUtils.interpolateSupport(self.apodizer,resolution,kind='linear')
        else:
            self.apodizer = 1.0
            self.path_apodizer = ''
            
        #----- MODAL BASIS FOR TELESCOPE ABERRATIONS
        if path_statModes!='' and ospath.isfile(path_statModes) == True and re.search(".fits",path_statModes)!=None:                
            statModes = np.asarray(fits.getdata(path_statModes))
            s1,s2,s3 = statModes.shape
            if s1 != s2: # mode on first dimension
                tmp = np.transpose(statModes,(1,2,0))
            else:
                tmp = statModes
                    
            self.nModes = tmp.shape[-1]
            self.statModes = np.zeros((resolution,resolution,self.nModes))
                
            for k in range(self.nModes):
                mode = FourierUtils.interpolateSupport(tmp[:,:,k],resolution,kind='linear')
                if pupilAngle !=0:
                    mode = rotate(mode,pupilAngle,reshape=False)
                self.statModes[:,:,k] = mode
        else:
            self.statModes = None
            self.nModes = 0
            self.path_statModes = ''
            
        if self.verbose:
            self
                   
            
    def __repr__(self):
        s = "___TELESCOPE___\n -------------------------------------------------------------------------------- \n"
        s += '. Aperture diameter \t:%.2fm \n'%(self.D)
        s += '. Central obstruction \t:%.2f%s \n'%(self.obsRatio*1e2,'%')
        s += '. Collecting area \t:%.2fm^2\n'%(self.area)
        s += '. Pupil resolution \t:%dX%d pixels'%(self.resolution,self.resolution)
        if self.path_pupil != '':
            s+= '\n. User-defined pupil at :\n ' + self.path_pupil
        if self.path_apodizer != '':
            s+= '\n. User-defined amplitude apodizer at :\n ' + self.path_apodizer
        if self.path_static_on != '':
            s+= '\n. User-defined on-axis static aberrations map at :\n ' + self.path_static_on
        if self.path_static_off != '':
            s+= '\n. User-defined off-axis static aberrations maps at :\n ' + self.path_static_off
        if self.path_statModes != '':
            s+= '\n. User-defined modes of static aberrations at :\n ' + self.path_statModes
        s = s +"\n--------------------------------------------------------------------------------\n"
        return s
        
    
    def displayPupil(self):
        """DISPLAYPUPIL Display the telescope pupil
        """
               
        import matplotlib.pyplot as plt
        plt.figure()
        plt.imshow(self.pupil)
        plt.xlabel("Pixels")
        plt.ylabel("Pixels")
        plt.title("Telescope pupil")
            
